import os
import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins
from . import mimportlib

__mlib__ = "AbcShutil"


class _interface(ABC):
    import shutil as local_shutil

    @property
    @abstractmethod
    def Error(self) -> local_shutil.Error:
        ...

    @property
    @abstractmethod
    def SpecialFileError(
        self,
    ) -> local_shutil.SpecialFileError:
        ...

    @property
    @abstractmethod
    def ReadError(self) -> local_shutil.ReadError:
        ...

    @property
    @abstractmethod
    def RegistryError(self) -> local_shutil.RegistryError:
        ...

    @property
    @abstractmethod
    def ExecError(self) -> local_shutil.ExecError:
        ...

    @property
    @abstractmethod
    def SameFileError(self) -> local_shutil.SameFileError:
        ...

    @property
    @abstractmethod
    def _GiveupOnFastCopy(self):
        return self.local_shutil._GiveupOnFastCopy

    @property
    @abstractmethod
    def _use_fd_functions(self):
        ...

    @property
    @abstractmethod
    def _ZLIB_SUPPORTED(self):
        ...

    @property
    @abstractmethod
    def _BZ2_SUPPORTED(self):
        ...

    @property
    @abstractmethod
    def _LZMA_SUPPORTED(self):
        ...

    @property
    @abstractmethod
    def _get_uid(self):
        ...

    @property
    @abstractmethod
    def _get_gid(self):
        ...

    @abstractmethod
    def copyfileobj(self, fsrc, fdst, length=0) -> None:
        ...

    @abstractmethod
    def copyfile(self, src, dst, *, follow_symlinks=True) -> typing.Any:
        ...

    @abstractmethod
    def copymode(self, src, dst, *, follow_symlinks=True) -> None:
        ...

    @abstractmethod
    def copystat(self, src, dst, *, follow_symlinks=True) -> None:
        ...

    @abstractmethod
    def copy(self, src, dst, *, follow_symlinks=True) -> typing.Any:
        ...

    @abstractmethod
    def copy2(self, src, dst, *, follow_symlinks=True) -> typing.Any:
        ...

    @abstractmethod
    def copytree(
        self,
        src,
        dst,
        symlinks=False,
        ignore=None,
        copy_function=copy2,
        ignore_dangling_symlinks=False,
        dirs_exist_ok=False,
    ) -> typing.Any:
        ...

    @abstractmethod
    def move(self, src, dst, copy_function=copy2) -> typing.Any:
        ...

    @abstractmethod
    def rmtree(self, path, ignore_errors=False, onerror=None, *, dir_fd=None) -> None:
        ...

    @abstractmethod
    def make_archive(
        self,
        base_name,
        format,
        root_dir=None,
        base_dir=None,
        verbose=0,
        dry_run=0,
        owner=None,
        group=None,
        logger=None,
    ) -> str:
        ...

    @abstractmethod
    def get_archive_formats(self) -> list[tuple[str, str]]:
        ...

    @abstractmethod
    def register_archive_format(
        self,
        name: str,
        function: typing.Callable[..., object],
        extra_args: typing.Sequence[tuple[str, typing.Any] | list],
        description: str = ...,
    ) -> None:
        ...

    @abstractmethod
    def unregister_archive_format(self, name: str) -> None:
        ...

    @abstractmethod
    def get_unpack_formats(self) -> list[tuple[str, list[str], str]]:
        ...

    @abstractmethod
    def register_unpack_format(
        self,
        name: str,
        extensions: list[str],
        function: typing.Callable[..., object],
        extra_args: typing.Sequence[tuple[str, typing.Any]],
        description: str = "",
    ) -> None:
        ...

    @abstractmethod
    def unregister_unpack_format(
        self,
        name: str,
        extensions: list[str],
        function: typing.Callable[..., object],
        extra_args: typing.Sequence[tuple[str, typing.Any]],
        description: str = ...,
    ) -> None:
        ...

    @abstractmethod
    def unpack_archive(self, filename, extract_dir=None, format=None) -> None:
        ...

    @abstractmethod
    def ignore_patterns(
        *patterns,
    ) -> typing.Callable[[typing.Any, list[str]], set[str]]:
        ...

    @abstractmethod
    def chown(self, path, user=None, group=None) -> None:
        ...

    @abstractmethod
    def which(self, cmd, mode=os.F_OK | os.X_OK, path=None) -> typing.Any:
        ...

    @abstractmethod
    def get_terminal_size(self, fallback=(80, 24)) -> os.terminal_size:
        ...

    # for test_shutil
    @abstractmethod
    def open(self):
        ...

    @abstractmethod
    def _copyfileobj_readinto(self):
        ...

    @abstractmethod
    def _destinsrc(self):
        ...


class AbcShutil(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mshutil"

    __all__ = [
        "copyfileobj",
        "copyfile",
        "copymode",
        "copystat",
        "copy",
        "copy2",
        "copytree",
        "move",
        "rmtree",
        "Error",
        "SpecialFileError",
        "ExecError",
        "make_archive",
        "get_archive_formats",
        "register_archive_format",
        "unregister_archive_format",
        "get_unpack_formats",
        "register_unpack_format",
        "unregister_unpack_format",
        "unpack_archive",
        "ignore_patterns",
        "chown",
        "which",
        "get_terminal_size",
        "SameFileError",
        "disk_usage",
    ]

    os = mimportlib.module_property("os")

    __delegated__ = {_interface: _interface.local_shutil}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_shutil, self)

        # tune module
        mod.disk_usage = self._myrrh_disk_usage

        mod._HAS_FCOPYFILE = False

        self.__delegate__(_interface, mod)

    def _myrrh_disk_usage(self, path):
        """Return disk usage statistics about the given path.

        Returned value is a named tuple with attributes 'total', 'used' and
        'free', which are the amount of total, used and free space, in bytes.
        """
        _ntuple_diskusage = self._ntuple_diskusage
        os = self.os

        st = os.statvfs(path)
        free = st.f_bavail * st.f_frsize
        total = st.f_blocks * st.f_frsize
        used = (st.f_blocks - st.f_bfree) * st.f_frsize
        return _ntuple_diskusage(total, used, free)

    def _myrrh__unpack_zipfile(self, filename, extract_dir):
        """Unpack zip `filename` to `extract_dir`"""
        raise NotImplementedError

    def _myrrh__make_zipfile(
        self,
        base_name,
        base_dir,
        verbose=0,
        dry_run=0,
        logger=None,
        owner=None,
        group=None,
        root_dir=None,
    ):
        """Create a zip file from all the files under 'base_dir'.

        The output zip file will be named 'base_name' + ".zip".  Returns the
        name of the output zip file.
        """
        raise NotImplementedError



import importlib
import typing

from myrrh.core.services import load_ext_group
from myrrh.core.services.entity import CoreProviderClass

from ._iprovider import IProvider


class ProviderRegistry:
    providers: dict[str, typing.Type[IProvider]] = {}

    provider_sep = "_"
    provider_prefix = provider_sep.join(("myrrh", "provider"))

    __single__: typing.Optional["ProviderRegistry"] = None

    def __new__(cls):
        if cls.__single__ is None:
            cls.__single__ = super().__new__(cls)
        return cls.__single__

    def __getattr__(self, name) -> typing.Type[IProvider]:
        return self.get(name)

    def get(self, name: str) -> typing.Type[IProvider]:
        try:
            return self.providers[name]
        except KeyError:
            pass

        raise AttributeError(f'{name} is not a valid provider name, available providers: {", ".join(self.providers)}')

    def is_provider_dist(self, dist_name: str) -> bool:
        return len(dist_name) > len(self.provider_prefix) + 1 and dist_name.startswith(self.provider_prefix + self.provider_sep)

    def dist_name(self, provider_name: str) -> str:
        return self.provider_sep.join((self.provider_prefix, provider_name))

    def name_from_dist(self, dist_name: str) -> str:
        if self.is_provider_dist(dist_name):
            return dist_name[len(self.provider_prefix) + 1 :]
        return ""

    def get_version(self, provider_name: str) -> str:
        if provider_name not in self.providers:
            return ""

        provider = self.providers[provider_name]
        # look for __version__.py module inside provider
        try:
            version_module = importlib.import_module(f"{provider.__name__}.__version__")
        except (ModuleNotFoundError, ImportError):
            pass

        version = "0.0.0"
        for attr in ("version", "__version__"):
            try:
                version = getattr(version_module, attr)
                break
            except AttributeError:
                pass

        return version

    def _new_provider(self, name, provider_cls: typing.Type[IProvider]):
        assert issubclass(provider_cls, IProvider), f"{repr(provider_cls)} is not a valid provider implementation, provider must inherit from IProvider"

        self.providers[name] = CoreProviderClass(provider_cls, name)

    def register(self, module_name, provider_cls):
        _, _, provider_name = module_name.rpartition(".")

        self._new_provider(provider_name, provider_cls)


def register_provider(module_name: str, provider: str):
    module = importlib.import_module(module_name)
    provider_cls = getattr(module, provider)
    return ProviderRegistry().register(module_name, provider_cls)


load_ext_group("myrrh.provider.registry")



import os
import glob
import time
import signal
import logging
import collections
import functools

from traceback import format_exc

from myrrh import factory
from myrrh.core.services import cfg_init
from myrrh.core.services.logging import log
from myrrh.framework import Runtime
from myrrh.utils import mshlex
from myrrh.utils.filemode import filemode

from myrrh.framework.msh import madvsh

from myrrh.core.interfaces import IHost
from myrrh.core.services import groups

from ._bmy_internal import entities
from ._bmy_entity import BmyEntity
from ._bmy_exceptions import (
    BmyNotReady,
    BmyException,
    BmyMyrrhFailure,
    BmyInvalidParameter,
    BmyInvalidEid,
    BmyExecutionFailure,
)

__all__ = [
    "load",
    "save",
    "eids",
    "isgroup",
    "groupkeys",
    "groupvalues",
    "isbuilt",
    "current",
    "next",
    "previous",
    "entity",
    "select",
    "unselect",
    "new",
    "build",
    "debug",
    "system",
    "edit",
    "push",
    "get",
    "rm",
    "rmdir",
    "mkdir",
    "chdir",
    "lsdir",
    "info",
    "setinfo",
    "reboot",
    "boot",
    "halt",
    "snap",
    "resnap",
    "desnap",
    "fstat",
    "joinpath",
    "realpath",
    "basename",
    "dirname",
    "abspath",
    "read",
    "write",
    "cp",
    "cptree",
    "pwd",
    "get",
    "push",
    "transfer",
    "move",
    "execute",
    "execute_mem_c",
    "execute_mem_ce",
    "launch",
    "kill",
    "which",
    "csnap",
    "snaps",
    "resnap",
    "list_providers",
    "bmy_func",
]

FILE_EXT = cfg_init("myrrh_file_ext", ".emyrrh", section="myrrh.framework.bmy")
BMY_ASYNC = cfg_init("use_async_group_for_bmy", True, section="myrrh.framework.bmy")
EXEC_MEMORY_LEN = cfg_init("execute_memory_len", 200, section="myrrh.framework.bmy")

if BMY_ASYNC:
    _func = groups.myrrh_group_async
else:
    _func = groups.myrrh_group


def bmy_func(valid_eid_required=True, attr_name="eid", attr_type=None):
    def _(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            eid = kwargs.pop(attr_name, None)

            if valid_eid_required:
                eid = entities.current(eid)
                if not (eid):
                    raise BmyInvalidEid(eid=eid)

            if attr_type:
                eid = entities.get(eid=eid)
                if attr_type != "entity":
                    eid = getattr(eid, attr_type)

            try:
                if entities.isgroup(eid):
                    return _func(func)(*args, **kwargs, eid=groups.MyrrhGroup(keys=eid))

                kwargs[attr_name] = eid
                return func(*args, **kwargs)

            except BmyException as exc:
                exc.eid = exc.eid or str(hasattr(eid, "cfg") and eid.cfg.id or eid)
                exc.func = func.__name__
                raise
            except OSError:
                raise
            except Exception as exc:
                raise BmyExecutionFailure(eid, func.__name__, msg=str(exc)).with_traceback(exc.__traceback__) from None

        wrapper.__doc__ = "".join(
            [
                wrapper.__doc__ or "",
                """
                Notice: When the eid parameter is None the default entity is used.
                """,
            ]
        )

        wrapper.__annotations__["eid"] = str

        return wrapper

    return _


def eids() -> list[str]:
    """List available eids

    Returns:
        list[str]: A list containing all the 'eid' managed by bmy
    """
    return entities.eids


def isgroup(eid: str | tuple | groups.MyrrhGroup) -> bool:
    """Return true if eid is a group

    Args:
        eid (str | tuple | groups.MyrrhGroup): Single eid or group.

    Returns:
        bool: true if eid is a group
    """
    return entities.isgroup(eid)


def groupkeys(group: str | tuple | groups.MyrrhGroup) -> tuple:
    """Return a tuple containing eid or group keys.

    Args:
        group (str | tuple | groups.MyrrhGroup): Single eid or group.

    Returns:
        tuple: A tuple containing group keys
    """
    return entities.groupkeys(group)


def groupvalues(group: groups.MyrrhGroup) -> tuple:
    """Return a tuple with all the values contained in group.

    Args:
        group (groups.MyrrhGroup): Single eid or group.

    Returns:
        tuple:  A tuple containing group values
    """
    return entities.groupvalues(group)


def current(eid=None, ident=None):
    """Get the current entity id

    Args:
        eid (str|tuple, optional): Entity id. Defaults to None.
        ident (int, optional): thread ident. Defaults to None.

    Returns:
        str|tuple:  default entity id if eid is None else eid
    """
    return entities.current(eid, ident)


def next():
    """
    Go to next selected entity id
    """
    entities.history.goto_next()


def previous():
    """
    Return to previous selected entity id in history
    """
    entities.history.goto_previous()


def entity(eid: str | tuple | None = None) -> BmyEntity:
    """Returns the entity object associated with the eid

    Args:
        eid (str | tuple | None, optional): Single eid or group. Defaults to None.

    Returns:
        BmyEntity: entity object
    """
    return entities[current(eid)]


class _select:
    _prev_eid: str | None = None

    def __init__(self, args, eid, use_import=None):
        if args and eid is not None:
            raise ValueError("select does not support positional and eid keyword parameters")

        if len(args) == 1:
            eid = current(args[0])
        elif args:
            eid = current(args)
        else:
            eid = current(eid)

        self._eid = eid

        if not self._eid:
            # do nothing just return
            return
        use_import = isbuilt(eid) if use_import is None else use_import
        self._use_import = use_import
        self._prev_eid = entities.history.get_current()
        entities.eid = eid

    def __call__(self):
        if self._eid:
            return entity()

    def __enter__(self):
        if self._eid and self._use_import:
            import mlib

            try:
                mlib.mlib_push(self().system)
            except (BmyNotReady, OSError):
                raise

        if self._eid:
            return entity(self._eid)

    def __str__(self):
        return self._eid

    def __repr__(self):
        return "bmy.select<%s> at 0x%x" % (self._eid, id(self))

    def __exit__(self, _t=None, _v=None, _tb=None):
        if self._eid:
            entities.eid = self._prev_eid
            if self._use_import:
                import mlib

                mlib.mlib_pop()


def select(*args, eid=None, use_import=True):
    """
    Select the provided entity identifier as the default entity

    The default entity can be set temporary when the method is called using a with statement.
    If use_import is False, the mlib package is not available

    Args:
        eid (str): entity id
        use_import (bool): enable or disable mlib import (default:True)

    Returns:
        context manager: select context manager
    """
    return _select(args, eid, use_import)


def unselect():
    """
    Unselect default entity

    No more default entity selected in bmy
    """
    entities.eid = None


@bmy_func(attr_type="entity")
def isbuilt(*, eid: BmyEntity):
    """
    Checked build state of an entity

    Returns:
        bool: True if eid is already built else False

    """

    return eid.built


@bmy_func(valid_eid_required=False)
def new(path: str | list[str], eid, *, warehouse=[], pre=[], post=[], **kwargs):
    """
    Instantiates an entity


    Args:
        path (str): provider path
        eid (str): eid for the new entity
        warehouse (List) : list of warehouse items
        **kwargs: keyword arguments to pass to the provider

    Returns:
        str: eid of the newly created entity

    """
    if isinstance(path, str):
        paths = [path]

    settings = list()
    for p in paths:
        settings.append({"name": factory.Assembly.provider_name(p), **kwargs})

    supply = factory.Assembly.Supply(paths=paths, settings=settings, pre=pre, post=post)  # type: ignore[call-arg]
    assembly = factory.Assembly(supply, warehouse=warehouse)

    try:
        return entities.append(assembly, eid)
    except BmyException as e:
        e.func = "new"
        raise
    except (RuntimeError, Exception) as e:
        raise BmyMyrrhFailure(func="new", msg="new entity failure") from e


@bmy_func(attr_type="entity")
def save(path=None, full=False, *, eid):
    if os.path.isdir(path):
        path = os.path.join(path, "".join((eid.eid, FILE_EXT)))

    if not path:
        path = "".join(eid, ".emyyrh")

    with open(path, "w") as f:
        f.write(eid._assembly.fromEntity(eid, only_predefined=not full).json(indent=2))


def load(path=None, eid=None):
    f"""
    Loads an entity information file ('{FILE_EXT}' file)

    This function is similar to:func:`bmy.new` excepts that the list of predefined warehouse is loaded from an emyrrh file

    Args:
        path (str|list|tuple): path to entity "{FILE_EXT}" file
        eid (str|list|tuple): entity eid
    Returns:
        str|list: eid of the newly created entity

    """
    if not path:
        path = os.getcwd()

    if os.path.isdir(path):
        paths = glob.glob(os.path.join(path, f"*{FILE_EXT}"))
        path = []
    else:
        paths = [path]

    if not paths:
        raise BmyMyrrhFailure(func="load", msg="load failure no file found")

    if not isinstance(eid, (list, tuple)):
        eids = (eid,) + (None,) * (len(paths) - 1)
    else:
        eids = tuple(eid) + (None,) * (len(paths) - len(eid))

    results = list()
    for p, eid in zip(paths, eids):
        try:
            assembly = factory.Assembly.fromFile(p)
            if eid is None:
                eid, _, _ = os.path.basename(p).partition(".")

            results.append(entities.append(assembly, eid))

        except BmyException as e:
            e.func = "load"
            raise
        except Exception as e:
            log.debug("load entity failed with exception\n%s" % format_exc())
            raise BmyMyrrhFailure(func="load", msg="load failure for %s: %s" % (p, e))

    return results if isinstance(path, (list, tuple)) or isinstance(eid, (list, tuple)) or len(results) > 1 else results[0]


@bmy_func()
def build(eid):
    """
    build an entity


    Args:
        eid(str): entity to builf

    Returns:
        str: eid of the newly created entity

    """
    try:
        entity(eid).build()
    except BmyException as e:
        e.func = "build"
        raise
    except Exception as e:
        log.debug("build entity failed with exception\n%s" % format_exc())
        raise BmyMyrrhFailure(msg="%s" % e)

    return eid


def debug(level="?"):
    """
    Set/unset Myrrh myrrh.logging

    Args:
        level (str): log level (ie: DEBUG, WARNING, INFO, ERROR, CRITICAL)

    Returns:
        int: current level or None if log is disabled

    """
    if level == "?":
        return log.level

    log.disabled = not level

    if log.disabled:
        return None

    level = logging.getLevelName(level.upper()) if hasattr(level, "upper") else level

    try:
        log.setLevel(level)
    except (ValueError, TypeError) as e:
        raise BmyInvalidParameter(msg='"%s" is not a valid level' % level) from e

    return log.level


@bmy_func()
def system(cmd, *, eid):
    """
    Executes a shell command on the entity

    Args:
        cmd (str): command line to execute on the entity
        eid (str): entity id

    Returns:
        int: cmd error code
    """
    with select(eid):
        from mlib.py import os

    return os.system(cmd)


@bmy_func()
def execute(
    cmd,
    *,
    path: str | None = None,
    count: int | None = None,
    duration: float | None = None,
    interval: float | None = None,
    ttl: float | None = None,
    timeout: float | None = None,
    poll: float | None = None,
    raiseonttl: bool = False,
    raiseontimeout: bool = True,
    executein: bool = True,
    eid=None,
):
    """Run a command on an entity.

    This function supports many arguments to manage repeat and duration cases.
    Like python Popen, the first 'execute' argument should be a sequence of program arguments or a single string.

    When executein is True, the command is immediately called. Otherwise, the command is executed during iteration calls.

    Args:
        cmd (str|list): shell command line to execute on the entity or executable arguments
        path (str | None, optional): executable absolute path. Defaults to None.
        count (int | None, optional): repeat the execution 'count' times. Defaults to None.
        duration (float | None, optional): repeat the execution until 'duration' is raised. Defaults to None.
        interval (float | None, optional): interval between two executions. Defaults to None.
        ttl (float | None, optional): time to live of one execution. Defaults to None.
        timeout (float | None, optional): global execution timeout. Defaults to None.
        poll (float | None, optional): _description_. Defaults to None.
        raiseonttl (bool, optional): raise if ttl. Defaults to False.
        raiseontimeout (bool, optional): raise if timeout. Defaults to True.
        executein (bool, optional): execute immediately. Defaults to True.
        eid (str|tuple): eid of the entity to use for the command execution.

    Returns:
        iterator on execution results
    """
    with select(eid):
        from mlib.sh import advsh

    exe = advsh.execute(
        cmd,
        path=path,
        count=count,
        during=duration,
        interval=interval,
        ttl=ttl,
        timeout=timeout,
        poll=poll,
        raiseonttl=raiseonttl,
        raiseontimeout=raiseontimeout,
    )

    execute.mem.append(exe)

    if executein:
        for _ in exe:
            pass

    return exe


execute.mem = collections.deque(maxlen=EXEC_MEMORY_LEN)
execute.TimeoutExpired = madvsh.TimeoutExpired
execute.TTLExpired = madvsh.TTLExpired


def execute_mem_c():
    """clear execution memory"""
    execute.mem.clear()


def execute_mem_ce(pid: int):
    """clear one element from execution memory

    Args:
        pid (int): pid of the element to clear

    Raises:
        ValueError: when pid not found
    """
    for exe in execute.mem:
        for exe_info in exe.calls:
            if exe_info.proc.pid == pid:
                break
        else:
            continue

        break
    else:
        raise ValueError(f"pid {pid} not found")

    exe.remove(exe_info)

    if not exe.ncalls:
        execute.mem.remove(exe)


@bmy_func()
def which(path, wdir=None, *, eid):
    """
    Returns the path to an executable which would be run if the given path was called.

    Args:
        path (str): executable name
        wdir (str): working directory (default=Not set)

        eid (str): entity id

    Returns:
        str: absolute path if executable is found else None
    """
    with select(eid):
        from mlib.py import shutil, os

    if wdir:
        wdir = os.pathsep.join([wdir, os.environ.get("PATH", os.defpath)])

    path = shutil.which(path, path=wdir)

    if path:
        path = os.path.abspath(path)

    return path


@bmy_func(attr_type="entity")
def launch(cmd, wdir=None, *, eid: BmyEntity):
    """
    Launch a process on entity, and returns its pid

    The first argument should be a sequence of program arguments or else a single string.
    By default, the path of the program to execute is the first item in cmd if cmd is a sequence.

    Note:
        If the path of the program to execute is not an absolute path, a request is emit to the entity to determine its absolute path, if not found
        the cmd is executed as it is.

    Args:
        cmd (str,list): command to execute
        wdir (str,list): process working directory (default=Not set)

        eid (str): entity id

    Returns:
        int: pid of the new created process, 0 on failure

    """
    if not isinstance(cmd, (list, tuple)):
        cmd = mshlex.split(cmd, posix=eid.cfg.system.os == "posix")

    relcmd, params = cmd[0], cmd[1:]

    abscmd = None
    if not eid.runtime.myrrh_os.isabs(relcmd):
        try:
            abscmd = which(relcmd, wdir, eid=eid)
            log.debug('relative cmd path "%s" set to "%s"' % (relcmd, abscmd))
        except BmyException:
            pass
    else:
        abscmd = relcmd

    if not abscmd:
        log.debug('relative cmd path "%s" can not be resolve' % relcmd)
        abscmd = relcmd

    cmd = [eid.runtime.myrrh_os.fsencode(a) for a in [abscmd, *params]]
    wdir = None if wdir is None else eid.runtime.myrrh_os.fsencode(wdir)

    return eid.system.shell.spawn(cmd, working_dir=wdir)


@bmy_func(attr_type="entity")
def kill(pid, sig=signal.SIGTERM, *, eid: BmyEntity):
    """
    Send a signal to process

    Args:
        pid (int): process pid
        sig (int): signal to send (default: SIGTERM)

        eid (str): entity id
    """
    eid.system.shell.signal(pid, sig)


@bmy_func()
def edit(path, nl=False, edit=None, *, eid):
    """
    Edit a file from an entity

    Load a file from an entity and copy it to a temporary file on the localhost.
    After edition, the file is uploaded on the entity.
    User can specify the editor path to be called for file edition.
    after closing the file is updated on the entity.

    Args:
        path (str): path to the file to edit on the entity
        nl (bool): convert new line (default: False)
        edit (callable): this function is called with the temporary file path

        eid (str): entity id
    """

    def _edit(tmpfile):
        editor = edit if isinstance(edit, str) else os.getenv("Bmy_EDITOR", "notepad") if os.name == "nt" else os.getenv("Bmy_EDITOR", "xdg-open")

        os.system('""%s" "%s""' % (editor, tmpfile))

    import tempfile

    with select(eid):
        from mlib.py import os as eos

    filename = eos.path.basename(path)
    _, ext = eos.path.splitext(path)
    _edit = _edit if isinstance(edit, str) or edit is None else edit

    nl = None if nl else ""
    tmpfile = None
    try:
        with tempfile.NamedTemporaryFile(prefix=filename, suffix=ext, delete=False, newline=nl) as tmp:
            tmpfile = tmp.name
            eos.myrrh_syscall.stream_in(eos.fsencode(path), tmp)

        import click

        click.launch(tmpfile)

        with open(tmpfile, "rb") as tmp:
            eos.myrrh_syscall.stream_out(eos.fsencode(path), tmp)

    finally:
        if tmpfile:
            os.remove(tmpfile)


@bmy_func()
def push(src_path, dest_path, *, eid: str, chunk_size=None):
    """
    Upload local files to an entity (src_path may contain simple shell-style wildcards)

    Args:
        src_path (str): source local path (may be a directory. in such a case, full directory content is pushed on remote entity)
        dest_path (str): destination entity path

        eid (str): destination entity id

    Returns:
        tuple: (list of created dirs, list of uploaded files)

    """
    with select(eid):
        from mlib.fs import advfs
        from mlib.py import os

    chunk_size = advfs.CHUNK_SZ if chunk_size is None else chunk_size

    if os.local_os.path.isdir(src_path):
        return advfs.pushdir(src_path, dest_path, chunk_size=chunk_size)

    return list(), advfs.pushfile(src_path, dest_path)


@bmy_func()
def get(src_path, dest_path, *, eid: str, chunk_size=None):
    """
    Download a file from the selected entity to localhost

    Args:
        src_path (str): entity source path
        dest_path (str): local destination path

        eid (str): source entity id

    Returns:
        tuple: list of downloaded files
    """

    with select(eid):
        from mlib.fs import advfs
        from mlib.py import os

    chunk_size = advfs.CHUNK_SZ if chunk_size is None else chunk_size

    if os.path.isdir(src_path):
        return advfs.getdir(src_path, dest_path, chunk_size=chunk_size)

    return list(), advfs.getfile(src_path, dest_path)


@bmy_func()
def transfer(target_eid, src_path, dest_path, *, eid: str, chunk_size=None):
    """
    Transfer of files from a target entity to the selected entity


    Args:
        target_eid (str); destination entity id
        src_path (str): source entity path (directory or file)
        dest_path (str): destination local path

        eid (str): source entity id

    Returns:
        tuple[list[str],list[str]]: transferred dir list, transferred files list

    """

    with select(eid):
        from mlib.fs import advfs

    chunk_size = advfs.CHUNK_SZ if chunk_size is None else chunk_size

    return advfs.transfer(entities[target_eid], src_path, dest_path, chunk_size=chunk_size)


@bmy_func()
def fstat(path, follow=False, *, eid: str):
    """
    Get information about a file

    Args:
        path (str): filepath
        follow (bool): if True follows link, else get information of the link (default: False)

        eid (str): entity id

    Returns:
        dict(): a dictionary containing file information ('file', 'size', 'username', 'groupname', 'access', 'atime', 'mtime', 'ctime', 'stat')

    """
    with select(eid):
        from mlib.py import os

    st = os.stat(path) if follow else os.lstat(path)
    sz = st.st_size

    try:
        with select(eid):
            from mlib.py import pwd, grp
        username = pwd.getpwuid(st.st_uid).pw_name
        groupname = grp.getgrgid(st.st_gid).gr_name
    except Exception:
        username = "-"
        groupname = "-"

    for unit in ["", "Ki", "Gi"]:
        if abs(sz) < 1024:
            break
        sz /= 1024

    sz = "%.1f%sB" % (sz, unit if sz < 1024 else "T")

    return {
        "file": path,
        "size": sz,
        "username": username,
        "groupname": groupname,
        "access": filemode(st.st_mode),
        "atime": time.ctime(st.st_atime),
        "mtime": time.ctime(st.st_mtime),
        "ctime": time.ctime(st.st_ctime),
        "stat": st,
    }


@bmy_func()
def cp(from_path, to_path="", *, eid: str):
    """
    Copy file or directory on an entity

    Note:
        if the destination path exits, the file is overwritten

    Args:
        from_path (str): source file/folder path
        to_path (str): destination file/folder path

        eid (str): entity id
    """
    with select(eid):
        from mlib.fs import advfs
    advfs.copy(from_path, to_path)


@bmy_func()
def cptree(from_path, to_path, *, eid: str):
    """
    Deprecated use:func:`cp` instead
    """
    cp(from_path, to_path)


@bmy_func()
def move(from_path, to_path, *, eid: str):
    """
    Move file or directory on entity

    Args:
        from_path (str): source file/folder path
        to_path (str): destination file/folder path

        eid (str): entity id
    """
    with select(eid):
        from mlib.fs import advfs

    advfs.move(from_path, to_path)


@bmy_func()
def rm(path, *, eid: str):
    """
    Remove a file or a complete directory tree on an entity

    Path argument can use patterns according to the python mod:`glob` module

    Args:
        path (str): file path to delete

        eid (str): entity id
    """
    with select(eid):
        from mlib.fs import advfs

    advfs.rm(path)


@bmy_func()
def rmdir(path, force=False, *, eid: str):
    """
    deprecated use rm instead

    Remove a directory on an entity

    Path argument can use patterns according to the python mod:`glob` module

    Args:
        path (str): file path to delete
        force (bool): removes all directories and files in the specified directory

        eid (str): entity id
    """
    return rm(path, eid=eid)


@bmy_func()
def mkdir(path, *, eid: str):
    """
    Creates a directory on an entity  (creates all intermediate-level directories).

    Args:
        path (str): directory folder path

        eid (str): entity id
    """
    with select(eid):
        from mlib.py import os

    os.makedirs(path, exist_ok=True)


@bmy_func()
def chdir(path, *, eid: str) -> tuple[str, str]:
    """
    Change the current working directory on an entity.
    This may be useful in case of bmy.{execute,launch,lsdir,mkdir,info,...}  calls

    Args:
        path (str): new working directory path

        eid (str): entity id

    Returns:
        tuple(str, str): previous working directory, new working directory
    """
    with select(eid):
        from mlib.py import os

    cwd = os.getcwd()
    os.chdir(path)

    return cwd, os.getcwd()


@bmy_func(attr_type="runtime")
def pwd(*, eid: Runtime):
    """
    Return path of the current working directory

    Args:
        eid (str): entity id

    Returns:
        str: current working directory path
    """
    return eid.myrrh_os.getpath()


@bmy_func()
def lsdir(path=None, *, eid: str):
    """
    List entries in directories of an entity

    Paths argument can used patterns according to the python mod:`glob` module

    Args:
        path (str): folder path or pattern (default = current folder path)

        eid (str): entity id

    Returns:
        list[str]: list directory entries

    """
    with select(eid):
        from mlib.py import glob
        from mlib.py import os

    path = path or os.curdir

    if glob.has_magic(path):
        return glob.glob(path)

    return os.listdir(path)


def list_providers():
    """
    Return the list of installed provider
    """
    from myrrh.provider.registry import ProviderRegistry

    return list(ProviderRegistry.providers)


@bmy_func(attr_type="entity")
def setinfo(__category, *, eid: BmyEntity, **kwa):
    """
    Append information in the warehouse of the entity

    Args:
        __category (str): item category
        **kwargs : item data
        eid (str): entity id

    """

    from myrrh.warehouse import GenericItem

    item = GenericItem(type_=__category, **kwa)

    try:
        item.model_dump_json()
    except Exception:
        raise BmyInvalidParameter(eid=eid, msg="unable to serialize")

    eid.cfg.append(item)


@bmy_func(attr_type="entity")
def info(name=None, *, eid: BmyEntity):
    """
    Extract property/information on an entity

    Note:
        When name parameter is not set, the method returns the following default information: 'eid', 'manage address', 'location', 'os', 'system', 'host', 'vendor', 'working dir', 'warehouse'


    Args:
        name (list(str)|str): list of properties to retrieve or a property name

        eid (str): entity id

    Returns:
        Any : value of required property

    Raises:
        BmyInvalidParameter: when invalid parameter type is used

    """

    if name is not None and not isinstance(name, str) and not isinstance(name, (list, tuple)):
        raise BmyInvalidParameter(eid, 'invalid type for parameter "name"')

    infos = {
        "id": lambda: eid.cfg.id.id,
        "description": lambda: eid.cfg.system.description,
        "location": lambda: (eid.cfg.host.hostname, eid.cfg.host.loc),
        "os": lambda: eid.cfg.system.os,
        "services": lambda: eid._entity and eid._entity.provider.paths.get("0_") or list(),
        "catalog": lambda: eid._entity and eid._entity.provider.paths.get("1_") or list(),
        "cwd": lambda: eid.runtime.myrrh_os.getpath(),
        "warehouse": lambda: ",".join(eid.cfg.keys()),
    }

    entries = (name,) if isinstance(name, str) else name if isinstance(name, (list, tuple)) else tuple(infos) if name is None else ()

    info = collections.OrderedDict()

    for entry in entries:
        key, path = entry.split(".", 1) if "." in entry else (entry, "")

        if key not in infos:
            path = entry
            key = "warehouse"

        try:
            if key == "warehouse" and path:
                path = filter(None, path.split("."))
                cfg = eid.cfg
                for p in path:
                    cfg = cfg[p]
                info[entry] = cfg
            else:
                info[entry] = infos.get(entry, lambda: "na")()

        except Exception as e:
            info[entry] = f'entry "{entry}" can not be evaluated: %s' % e

    return info.get(name) if isinstance(name, str) else info


@bmy_func(attr_type="host")
def reboot(wait=False, force=False, *, eid: IHost):
    """
    Reboot entity

    Args:
        wait (bool): wait reboot complete.
        force (bool): hard reboot

        eid (str): entity id

    """
    eid.state.reset(wait=wait, force=force)


@bmy_func(attr_type="host")
def boot(wait=False, *, eid: IHost):
    """
    Boot an entity

    Args:
        wait (bool): wait boot complete
    """
    eid.state.start(wait=wait)


@bmy_func(attr_type="host")
def halt(force=False, wait=False, *, eid: IHost):
    """
    Halt entity

    Args:
        wait (bool): wait halt complete
        force (bool): hard shutdown

        eid (str): entity id

    """
    eid.state.stop(force=force, wait=wait)


@bmy_func(attr_type="host")
def snap(name, *, eid: IHost):
    """
    Take a snapshot of an entity

    Note:
        By default, perform a snapshot of the entity current state, if snapshot capability offered (by the manager)

    Args:
        name(str): name of the snapshot

        eid (str): entity id

    """
    eid.snap.new(name)


@bmy_func(attr_type="host")
def resnap(name, *, eid: IHost):
    """
    Restore a snapshot

    Note:
        If an entity does not support live snapshot, the new state of the entity is halted after being restored. (docker <v19.x)

    Args:
        name (str): name of the snapshot to restore

        eid (str): entity id

    """

    eid.snap.restore(name)


@bmy_func(attr_type="host")
def desnap(name, *, eid: IHost):
    """
    Delete a snapshot

    Args:
        name (str): name of the snapshot

        eid (str): entity id

    """
    eid.snap.delete(name)


@bmy_func(attr_type="host")
def snaps(eid: IHost):
    """
    Returns a list of available snapshots

    Args:
        eid (str): entity id

    Returns:
        list(): names of available snapshot

    """
    return eid.snap.names()


@bmy_func(attr_type="host")
def csnap(eid: IHost):
    """
    Return name of current snapshot

    Args:
        eid (str): entity id

    Returns:
        str: name of current snapshot
    """
    return eid.snap.current()


@bmy_func(attr_type="runtime")
def joinpath(*args, eid: Runtime):
    "Join one or more path warehouse"
    if not args:
        raise BmyInvalidParameter("a path is required")

    return eid.myrrh_os.fscast(args[0])(eid.myrrh_os.joinpath(*(eid.myrrh_os.f(a) for a in args)))


@bmy_func(attr_type="entity")
def realpath(path, eid: Runtime):
    "Return the canonical path of the specified filename"
    with select(eid):
        from mlib.py import os

    return os.realpath(path)


@bmy_func(attr_type="runtime")
def basename(path, eid: Runtime):
    "Return the base name part of a path"
    return eid.myrrh_os.fscast(path)(eid.myrrh_os.basename(eid.myrrh_os.f(path)))


@bmy_func(attr_type="runtime")
def dirname(path, eid: Runtime):
    "Return the directory name of the specified path"
    return eid.myrrh_os.fscast(path)(eid.myrrh_os.dirname(eid.myrrh_os.f(path)))


@bmy_func(attr_type="runtime")
def abspath(path, eid: Runtime):
    "Return the absolute path of the specified filename"
    return eid.myrrh_os.fscast(path)(eid.myrrh_os.getpathb(eid.myrrh_os.f(path)))


@bmy_func(attr_type="runtime")
def write(path, data=b"", *, encoding=None, errors=None, eid: Runtime):
    "Write data to file (!! overwrite existing one)"
    import io

    if isinstance(data, str):
        if not encoding:
            encoding = eid.myrrh_os.defaultencoding()

        data = data.encode(encoding, errors)

    with io.BytesIO(data) as e:
        eid.myrrh_syscall.stream_out(eid.myrrh_os.fsencode(path), e)


@bmy_func(attr_type="runtime")
def read(path, *, binary=False, encoding=None, errors=None, eid: Runtime):
    "Read file data and return it"
    import io

    if not encoding:
        encoding = eid.myrrh_os.defaultencoding
    if not errors:
        errors = "strict"

    with io.BytesIO() as data:
        eid.myrrh_syscall.stream_in(eid.myrrh_os.fsencode(path), data)
        return data.getvalue() if binary else data.getvalue().decode(encoding, errors)



import functools

from myrrh.core.services.system import ExecutionFailureCauseRVal

from myrrh.framework.mpython._mosenv import AbcOsEnv

__mlib__ = "OsEnv"


class OsEnv(AbcOsEnv):
    name = "posix"
    linesep = "\n"

    def __init__(self, system):
        super(AbcOsEnv, self).__init__(system=system)
        self._path = None

    def putenv(self, varname, value):
        _, err, rval = self.myrrh_os.cmdb(
            b'%(echo)s "export %(varname)s=%(value)s" >> $HOME/.profile',
            varname=self.myrrh_os.sh_escape_bytes(varname),
            value=self.myrrh_os.sh_escape_bytes(value),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        _, err, rval = self.myrrh_os.cmdb(
            b'%(echo)s "export %(varname)s=%(value)s" >> $HOME/.bashrc',
            varname=self.myrrh_os.sh_escape_bytes(varname),
            value=self.myrrh_os.sh_escape_bytes(value),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def unsetenv(self, varname):
        _, err, rval = self.myrrh_os.cmdb(
            rb'%(sed)s -i /"%(varname)s=.*"/d $HOME/.profile',
            varname=self.myrrh_os.sh_escape_bytes(varname),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        _, err, rval = self.myrrh_os.cmdb(
            rb'%(sed)s -i /"%(varname)s=.*"/d $HOME/.bashrc',
            varname=self.myrrh_os.sh_escape_bytes(varname),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def gethome(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s $HOME")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def gettemp(self):
        out, err, rval = self.myrrh_os.cmd(b"%(dirname)s `mktemp -u`")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def getlogin(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s $USER")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def geteuid(self):
        return self.cfg.system.uid or 0

    def getegid(self):
        return self.cfg.system.gid or 0

    def getuid(self):
        return self.geteuid()

    def getgid(self):
        return self.getegid()

    def getgroups(self):
        return self.cfg.system.groups or [0]

    @functools.cache
    def cpu_count(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(cat)s  /proc/cpuinfo")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        s = 0
        c = 0
        while s > -1:
            c += 1
            s = out.find(b"processor", s + len(b"processor"))
        return None if c == 0 else c



# flake8: noqa: F401

import re

from click.types import BOOL, FLOAT, INT, UNPROCESSED, UUID, ParamType, STRING, Choice


class RegExType(ParamType):
    def __init__(self, reg):
        self.creg = re.compile(reg)

    def convert(self, value, param, ctx):
        result = self.creg.match(value)
        if not result:
            self.fail('"%s" invalid expression' % value, param, ctx)

        return result



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from myrrh.framework.mpython import mbuiltins

__mlib__ = "AbcOsHelper"


class _interface(ABC):
    import test.support.os_helper as local_os_helper

    @property
    @abstractmethod
    def EnvironmentVarGuard(self) -> local_os_helper.EnvironmentVarGuard:
        ...

    @property
    @abstractmethod
    def FakePath(self) -> local_os_helper.FakePath:
        ...

    @property
    @abstractmethod
    def TESTFN(self) -> str:
        ...

    @property
    @abstractmethod
    def TESTFN_ASCII(self) -> str:
        ...

    @abstractmethod
    def skip_if_dac_override(self, test) -> typing.Any:
        ...

    @abstractmethod
    def skip_unless_xattr(self, test) -> typing.Any:
        ...

    @abstractmethod
    def skip_unless_dac_override(self, test) -> typing.Any:
        ...

    @abstractmethod
    def skip_unless_working_chmod(self, test) -> typing.Any:
        ...

    @abstractmethod
    def skip_unless_symlink(self, test) -> typing.Any:
        ...

    @abstractmethod
    def can_symlink(self, test) -> typing.Any:
        ...

    @abstractmethod
    def create_empty_file(self, filename) -> None:
        ...

    @abstractmethod
    def change_cwd(self, path, quiet: bool = ...) -> typing.ContextManager[str]:
        ...

    @abstractmethod
    def unlink(self, filename) -> None:
        ...

    @abstractmethod
    def make_bad_fd(self) -> int:
        ...

    @abstractmethod
    def rmtree(self, path) -> None:
        ...

    @abstractmethod
    def temp_dir(self, path=None, quiet=False) -> typing.Any:
        ...

    @abstractmethod
    def temp_cwd(self, name="tempcwd", quiet=False) -> typing.Any:
        ...

    @abstractmethod
    def rmdir(self, dirname) -> typing.Any:
        ...


class AbcOsHelper(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mtest.msupport.mos_helper"

    __delegated__ = {_interface: _interface.local_os_helper}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_os_helper, self)
        self.__delegate__(_interface, mod)

        # for test purpose
        self._mod = mod



import time
import typing


class MTimer:
    class DelayExpired(TimeoutError):
        timer: typing.Optional["MTimer"] = None

    class CountExpired(TimeoutError):
        timer: typing.Optional["MTimer"] = None

    INFINITE = -1

    def __init__(self, timeout=None, count=None, step=None, raise_on_expired=False):
        """
        Create simple synchronous timer

        :param int timeout: timer expiration in second
        :param in count: timer expiration in discrete time count
        """
        self._count = count
        self._timeout = timeout
        self._step_sleep_time = step
        self._raise_on_expired = raise_on_expired

        self._idle_endtime = 0

        self.reset()

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, *_a, **_k):
        self.stop()

    def __bool__(self):
        return self.idle

    @property
    def _timeleft(self):
        return self._timeout is None or self._timeout and max(0, self._timeout - self.duration)

    @property
    def _countleft(self):
        return self._count is None or self._count == self.INFINITE or self._count and max(0, self._count - self._nbstep)

    @property
    def duration(self):
        _time = self._endtime or self._ctime
        return _time - self._starttime

    @property
    def count(self):
        return self._count

    @property
    def timeout(self):
        return self._timeout

    @property
    def step(self):
        return self._step_sleep_time

    @property
    def cstep(self):
        return self._nbstep

    @property
    def starttime(self):
        return self._starttime

    @property
    def endtime(self):
        return self._endtime

    @property
    def expired(self):
        return self._nbstep > 0 and (self._endtime or not self._timeleft or not self._countleft)

    @property
    def expired_on_count(self):
        return self._count and not self._countleft

    @property
    def expired_on_timeout(self):
        return self._timeout and not self._timeleft

    @property
    def idle(self):
        self._nbstep += 1

        if self.expired:
            self.stop()
            if self._raise_on_expired:
                expired = TimeoutError()
                if not self._count and not self._timeout:
                    expired = self.CountExpired()
                elif self.expired_on_timeout:
                    expired = self.DelayExpired()
                elif self.expired_on_count:
                    expired = self.CountExpired()
                else:
                    expired = TimeoutError()
                expired.timer = self
                raise expired
            else:
                return False
        if self._step_sleep_time:
            time.sleep(self.sleep_time)

        self._idle_endtime = self._ctime

        return True

    @property
    def sleep_time(self):
        if self._step_sleep_time:
            return min(self._step_sleep_time, self._timeleft) if self._timeout else self._step_sleep_time

    @property
    def timeleft(self):
        if self._timeout is not None:
            return self._timeleft

        # None

    @property
    def countleft(self):
        if self._count is not None:
            return self._countleft

        # None

    def reset(self):
        self._starttime = self._ctime
        self._nbstep = 0
        self._endtime = None

    def stop(self):
        self._endtime = self._ctime

    @property
    def _ctime(self):
        return time.monotonic()


class MVectorTimer(MTimer):
    def __init__(self, vectors, *a, **kw):
        super().__init__(*a, **kw)
        self._vectors = list(vectors)
        self._stamp = 0

    @property
    def sleep_time(self):
        vector = int((self.duration - self._stamp) / self._step_sleep_time)
        wait = self._vectors[min(len(self._vectors) - 1, vector)]
        return wait

    def stamp(self):
        self._stamp = self.duration


class MBurstTimer(MVectorTimer):
    _VECTOR = [0.01] * 10 + [0.1] * 10 + [0.2] * 30 + [0.3] * 50 + [0.5] * 500 + [1]
    _STEP = 0.1

    def __init__(self, vectors=None, timeout=None, count=None, step=None, raise_on_expired=False):
        super().__init__(
            vectors=vectors and list(vectors) or self._VECTOR,
            timeout=timeout,
            count=count,
            step=step or self._STEP,
            raise_on_expired=raise_on_expired,
        )



from ...interfaces import IRuntimeObject, ICoreService

__all__ = ("RuntimeObjectProxy", "RuntimeObject")


class RuntimeObject(IRuntimeObject):
    __slots__ = ("epath", "ename", "ehandle", "eref")

    epath: bytes
    ename: bytes
    ehandle: int
    eref: str

    closed = False

    def __init__(self, handle: int, path: bytes, name: bytes, service: ICoreService | None):
        self.ehandle: int = handle
        self.service = service
        self.epath = path
        self.ename = name

        if service:
            self.eref = "/".join((service.eref(), str(self.ehandle)))
        else:
            self.eref = f"./{id(self)}"

    def __del__(self):
        self.close()

    def close(self):
        self.closed = True


class RuntimeObjectProxy(RuntimeObject):
    def __init__(self, obj, stream):
        handle = getattr(obj, "fileno", lambda: -777)()
        name = getattr(obj, "name", b"")
        path = getattr(obj, "path", b"")

        super().__init__(handle, path, name, stream)

        self.obj = obj

    def __getattr__(self, name):
        return getattr(self.obj, name)



"""Filename matching with shell patterns.

fnmatch(FILENAME, PATTERN) matches according to the local convention.
fnmatchcase(FILENAME, PATTERN) always takes case in account.

The functions operate by translating the pattern into a regular
expression.  They cache the compiled regular expressions for speed.

The function translate(PATTERN) returns a regular expression
corresponding to PATTERN.  (It does not compile it.)
"""
from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate
from . import mbuiltins

__mlib__ = "AbcFnmatch"


class _interface(ABC):
    import fnmatch as local_fnmatch

    @abstractmethod
    def filter(self, names, pat) -> list:
        ...

    @abstractmethod
    def fnmatch(self, name, pat) -> bool:
        ...

    @abstractmethod
    def fnmatchcase(self, pathname) -> bool:
        ...

    @abstractmethod
    def translate(self, pathname) -> str:
        ...


class AbcFnmatch(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mfnmatch"

    __all__ = ["filter", "fnmatch", "fnmatchcase", "translate"]

    __doc__ = _interface.local_fnmatch.__doc__

    __delegated__ = {_interface: _interface.local_fnmatch}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_fnmatch, self)

        self.__delegate__(_interface, mod)



import pydantic
import typing

SettingsT = typing.TypeVar("SettingsT")


class Settings(pydantic.BaseModel):
    def attrs(self) -> dict[str, typing.Any]:
        return super().model_dump(
            exclude={
                "name",
            }
        )


class Supply(pydantic.BaseModel, typing.Generic[SettingsT]):
    settings: list[SettingsT] = pydantic.Field(default_factory=list)

    paths: list[str]

    pre: list = pydantic.Field(default_factory=list)
    post: list = pydantic.Field(default_factory=list)



# flake8: noqa: F401

from myrrh.framework.bmy import __all__ as __all__
from myrrh.framework.bmy import __all_global__ as __all_global__
from myrrh.framework.bmy import __all_manage__ as __all_manage__
from myrrh.framework.bmy import __all_state__ as __all_state__
from myrrh.framework.bmy import __all_fs__ as __all_fs__
from myrrh.framework.bmy import __all_exec__ as __all_exec__
from myrrh.framework.bmy import __all_snaps__ as __all_snaps__
from myrrh.framework.bmy import __all_except__ as __all_except__
from myrrh.framework.bmy import __all_loading__ as __all_loading__

from myrrh.framework.bmy import *



import os
import sys
import threading
import typing

from myrrh.core.interfaces import ICoreStreamService

from ...interfaces import (
    IFileInOutStream,
    IFileInStream,
    IFileOutStream,
    IInStream,
    IOutStream,
    ITask,
    IRuntimeObject,
    Stat,
    ABCDelegation,
)

from ....provider import Whence, StatField

from ._objects import RuntimeObject
from ._pipe import _Lock, Buffer

__all__ = (
    "FileStream",
    "FileInStreamAsync",
    "FileOutStreamAsync",
    "FileInOutStream",
    "FileInStream",
    "FileOutStream",
    "InPipe",
    "OutPipe",
    "PipeInStreamAsync",
    "PipeOutStreamAsync",
    "InputIo",
)


class FileStream(RuntimeObject, IFileInOutStream):
    _eot = False
    service: ICoreStreamService

    def __init__(self, handle: int, path: bytes, name: bytes, service: ICoreStreamService):
        super().__init__(handle, path, name, service)

    @property
    def protocol(self):
        return self.service.protocol

    @property
    def eot(self):
        return self._eot

    def send_eot(self, eot, id=None):
        self._eot = eot

    def read(self, nbytes: int | None = 0, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        if nbytes is None:
            return self.service.readchunk(self.ehandle, extras=extras)
        if nbytes == 0:
            return self.service.readall(self.ehandle, extras=extras)
        return self.service.read(self.ehandle, nbytes, extras=extras)

    def write(self, data: bytes, *, extras: dict[str, typing.Any] | None = None):
        return self.service.write(self.ehandle, data, extras=extras)

    def close(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        if not self.closed:
            try:
                self.service.close(self.ehandle, extras=extras)
            except OSError:
                ...
            RuntimeObject.close(self)

    def seek(
        self,
        pos: int,
        whence: int = os.SEEK_SET,
        *,
        extras: dict[str, typing.Any] | None = None,
    ) -> int:
        try:
            whence = {
                os.SEEK_CUR: Whence.SEEK_CUR,
                os.SEEK_SET: Whence.SEEK_SET,
                os.SEEK_END: Whence.SEEK_END,
            }[whence].value

            return self.service.seek(self.ehandle, pos, whence, extras=extras)

        except KeyError:
            pass

        raise ValueError(f"Unknown whence method: {whence}")

    def sync(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        return self.service.sync(self.ehandle, extras=extras)

    def flush(self, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        return self.service.flush(self.ehandle, extras=extras)

    def truncate(self, length: int, *, extras: dict[str, typing.Any] | None = None) -> None:
        return self.service.truncate(self.ehandle, length, extras=extras)

    def stat(self, *, extras: dict[str, typing.Any] | None = None) -> Stat:
        return Stat(**self.service.stat(self.ehandle, fields=StatField.FILE.value))


class FileInStream(IFileInStream, ABCDelegation):
    __delegated__ = (IFileInStream, IRuntimeObject)

    def __init__(self, stream: FileStream):
        self.__delegate__(IFileInStream, stream)
        self.__delegate__(IRuntimeObject, stream)


class FileOutStream(IFileOutStream, ABCDelegation):
    __delegated__ = (IFileOutStream, IRuntimeObject)

    def __init__(self, stream: FileStream):
        self.__delegate__(IFileOutStream, stream)
        self.__delegate__(IRuntimeObject, stream)


class FileInOutStream(IFileInOutStream, ABCDelegation):
    __delegated__ = (IFileInOutStream, IRuntimeObject)

    def __init__(self, stream: IFileInOutStream):
        self.__delegate__(IFileInOutStream, stream)
        self.__delegate__(IRuntimeObject, stream)


class _Pipe(IRuntimeObject, ABCDelegation):
    __delegated__ = (IRuntimeObject,)

    def __init__(self, buffer: Buffer | None = None):
        self.buffer = buffer or Buffer()
        self._eot = False
        self.__delegate__(IRuntimeObject, self.buffer)

    def send_eot(self, val: bool, id_: int | None = None):
        if self._eot != val:
            self.buffer.send_eot(val, id_ or id(self))
            self._eot = val


class InPipe(_Pipe, IInStream, ABCDelegation):
    __delegated__ = (IInStream,)

    def __init__(self, buffer: Buffer | None = None):
        super().__init__(buffer)
        self.__delegate__(IInStream, self.buffer)


class OutPipe(_Pipe, IOutStream, ABCDelegation):
    __delegated__ = (IOutStream,)

    def __init__(self, buffer: Buffer | None = None):
        super().__init__(buffer)
        self.__delegate__(IOutStream, self.buffer)


class PipeInStreamAsync(ITask, ABCDelegation):
    def __init__(self, stream: FileStream, pipe: InPipe):
        self.data = bytearray()
        self.din_lock = _Lock(threading.RLock())
        self.pipe = pipe
        self.stream = stream
        self.exc = None
        self.pipe.send_eot(False)

    def __str__(self):
        return f"<{self.stream.eref}({self.stream.ename})"

    def _eot(self, eot):
        with self.din_lock:
            if self.pipe.eot != eot:
                self.pipe.send_eot(eot)
                self.din_lock.notify_all()

                if eot:
                    self.close()

    def task(self):
        try:
            with self.din_lock:
                if not self.stream.closed:
                    self.data = bytearray(self.stream.read(nbytes=None))
                    if not len(self.data):
                        self._eot(True)

            with self.pipe.buffer.wr_lock:
                p = 0
                try:
                    while p < len(self.data):
                        p += self.pipe.buffer.write(self.data[p:])
                finally:
                    self.data = self.data[p:]

        except Exception as e:
            self.exc = e
            self._eot(True)
            raise

    def sync(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        if self.exc:
            raise self.exc

        with self.din_lock:
            self.stream.sync(extras=extras)

    def close(self, *, extras: dict[str, typing.Any] | None = None):
        try:
            with self.din_lock:
                try:
                    self.pipe.send_eot(True)
                    self.sync()
                except Exception:
                    pass
                finally:
                    self.pipe.close()

        finally:
            try:
                self.stream.close(extras=extras)
            except OSError:
                ...

            with self.din_lock:
                self.din_lock.notify_all()

    def terminated(self):
        return int(self.stream.closed) or None


class PipeOutStreamAsync(ITask, ABCDelegation):
    def __init__(self, stream: FileStream, pipe: OutPipe):
        self.data = bytearray()
        self.dout_lock = _Lock(threading.RLock())
        self.pipe = pipe
        self.stream = stream
        self.exc = None

    def __str__(self):
        return f">{self.stream.eref}({self.stream.ename})"

    def _eot(self):
        self.close()

    def task(self):
        try:
            self.data = self.pipe.buffer.read()
            if not self.data:
                self._eot()
            else:
                self._write()
        except Exception as e:
            self.exc = e
            self._eot()
            raise

    def _write(self, *, extras: dict[str, typing.Any] | None = None):
        p = 0
        with self.dout_lock:
            try:
                while p < len(self.data):
                    p += self.stream.write(self.data[p:], extras=extras)
            finally:
                self.data = self.data[p:]

    def close(self, *, extras: dict[str, typing.Any] | None = None):
        try:
            with self.pipe.buffer.rd_lock:
                with self.dout_lock:
                    try:
                        self.pipe.send_eot(True)
                        self.sync()
                    except Exception:
                        pass
                    finally:
                        self.pipe.close()
        finally:
            try:
                self.stream.close(extras=extras)
            except OSError:
                pass

            with self.dout_lock:
                self.dout_lock.notify_all()

    def sync(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        if self.exc:
            raise self.exc

        with self.pipe.buffer.rd_lock:
            with self.dout_lock:
                self.data.extend(self.pipe.flush())
                assert self.pipe.buffer.rd_size == 0, self.pipe.buffer.rd_size
                self._write(extras=extras)

    def terminated(self):
        return int(self.stream.closed) or None


class FileInStreamAsync(PipeInStreamAsync, IFileInStream, ABCDelegation):
    __delegated__ = (IFileInStream,)

    def __init__(self, stream: FileStream, pipe: InPipe | None = None):
        super().__init__(stream, pipe or InPipe())
        self.__delegate__(IFileInStream, stream)
        self.tell = self.stream.seek(0, os.SEEK_CUR)
        self.pipe.send_eot(False)

    def _eot(self, eot):
        with self.din_lock:
            if self.pipe.eot != eot:
                self.pipe.send_eot(eot)
                self.din_lock.notify_all()

    def flush(self, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        with self.din_lock:
            with self.pipe.buffer.wr_lock:
                data = self.data
                self.data = bytearray()
                data.extend(self.pipe.flush())
                data.extend(self.stream.flush(extras=extras))
                self._eot(False)
            return data

    def seek(
        self,
        offset: int = 0,
        whence=os.SEEK_SET,
        *,
        timeout: float | None = None,
        extras: dict[str, typing.Any] | None = None,
    ) -> int:
        match whence:
            case os.SEEK_SET | os.SEEK_CUR:
                offset = offset + self.tell if whence == os.SEEK_CUR else offset

                if self.tell == offset:
                    ...
                else:
                    with self.din_lock:
                        self.flush()
                        self.tell = self.stream.seek(offset, os.SEEK_SET, extras=extras)
                        self._eot(False)

            case os.SEEK_END:
                with self.din_lock:
                    self.flush()
                    self.tell = self.stream.seek(offset, whence, extras=extras)
                    self._eot(False)

            case _:
                raise ValueError(whence)

        return self.tell

    def read(self, nbytes=None, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        if self.exc:
            raise self.exc

        data = self.pipe.read(nbytes)

        self.tell += len(data)

        return data


class FileOutStreamAsync(PipeOutStreamAsync, IFileOutStream, ABCDelegation):
    __delegated__ = (IFileOutStream,)

    def __init__(self, stream: FileStream, pipe: OutPipe | None = None):
        super().__init__(stream, pipe or OutPipe())
        self.__delegate__(IFileOutStream, stream)
        self.tell = self.stream.seek(0, os.SEEK_CUR)

    def _eot(self):
        pass  # no auto close

    def flush(self, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        if self.exc:
            raise self.exc

        with self.dout_lock:
            data = self.data
            self.data = bytearray()
            data.extend(self.stream.flush(extras=extras))

        return data

    def seek(
        self,
        offset: int = 0,
        whence=os.SEEK_SET,
        *,
        extras: dict[str, typing.Any] | None = None,
    ) -> int:
        if self.exc:
            raise self.exc

        match whence:
            case os.SEEK_SET | os.SEEK_CUR:
                offset = offset + self.tell if whence == os.SEEK_CUR else offset

                if self.tell != offset:
                    with self.dout_lock:
                        self.sync()
                        self.tell = self.stream.seek(offset, os.SEEK_SET, extras=extras)

            case os.SEEK_END:
                with self.dout_lock:
                    self.sync()
                    self.tell = self.stream.seek(offset, whence, extras=extras)

            case _:
                raise ValueError(whence)

        return self.tell

    def truncate(self, length: int, *, extras: dict[str, typing.Any] | None = None) -> None:
        if self.exc:
            raise self.exc

        with self.dout_lock:
            self.sync()
            self.stream.truncate(length, extras=extras)

    def write(
        self,
        data,
        *,
        extras: dict[str, typing.Any] | None = None,
        timeout: float | None = None,
    ) -> int:
        sz = 0

        if self.exc:
            raise self.exc

        try:
            sz += self.pipe.write(data)
        finally:
            self.tell += sz

        return sz


class InputIo(IInStream):
    ehandle = -1
    epath = b""
    eref = ""
    ename = b"stdin"

    @property
    def closed(self):
        return False

    @property
    def eot(self) -> bool:
        return False

    def send_eot(self, eot, id=None):
        return

    def sync(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        return

    def flush(self, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        return bytearray(b"")

    def stat(self, *, extras: dict[str, typing.Any] | None = None) -> Stat:
        st = os.stat(sys.stdin.fileno())
        return Stat._make(
            (
                st.st_mode,
                st.st_ino,
                st.st_dev,
                st.st_nlink,
                st.st_uid,
                st.st_gid,
                st.st_size,
                st.st_atime,
                st.st_mtime,
                st.st_ctime,
                0,
                None,
            )
        )

    def read(self, nbytes: int | None = None, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        import _winapi

        stdin = _winapi.GetStdHandle(_winapi.STD_INPUT_HANDLE)  # type: ignore[attr-defined]
        data, _, _ = _winapi.PeekNamedPipe(stdin)  # type: ignore[misc,attr-defined]

        return bytearray(data)

    def close(self):
        self.pinput.__exit__()



import abc
import enum
import typing

__all__ = (
    "IShellService",
    "IFileSystemService",
    "IStreamService",
    "IStateService",
    "ISnapService",
    "IInstanceService",
    "ServiceGroup",
    "Protocol",
    "IService",
    "Stat",
    "Wiring",
    "Whence",
    "StatField",
)


class Stat(typing.NamedTuple):
    st_mode: int = 0
    st_ino: int = 0
    st_dev: int = 0
    st_nlink: int = 0
    st_uid: int = 0
    st_gid: int = 0
    st_size: int = 0
    st_atime: int = 0
    st_mtime: int = 0
    st_ctime: int = 0
    st_pid: int | None = None
    st_status: int | None = None


class StatField(enum.Flag):
    MODE = 0x001
    INO = 0x002
    DEV = 0x004
    NLINK = 0x008
    UID = 0x010
    GID = 0x020
    SIZE = 0x040
    ATIME = 0x080
    MTIME = 0x100
    CTIME = 0x200
    PID = 0x400
    STATUS = 0x800
    FILE = MODE | INO | DEV | NLINK | UID | GID | SIZE | ATIME | MTIME | CTIME
    ALL = MODE | INO | DEV | NLINK | UID | GID | SIZE | ATIME | MTIME | CTIME | PID | STATUS


class Wiring(enum.Flag):
    OFF = 0
    IN = 0x01
    OUT = 0x02
    ERR = 0x04
    CREATE = 0x10
    RESET = 0x20
    INOUT = IN | OUT
    INOUTERR = IN | OUT | ERR


class Whence(enum.Enum):
    SEEK_SET = 0
    SEEK_CUR = 1
    SEEK_END = 2


class Protocol(enum.Enum):
    MYRRH = "m"
    POSIX = "posix"
    WINAPI = "winapi"
    VENDOR = "vendor"

    def __str__(self):
        return self._value_


class IService(abc.ABC):
    category: str
    name: str
    protocol: Protocol | str


T = typing.TypeVar("T")


class ServiceGroup(enum.Enum):
    vendor = abc.ABCMeta("IServiceVendor", (IService,), dict())
    system = abc.ABCMeta("IServiceSystem", (IService,), dict())
    host = abc.ABCMeta("IServiceHost", (IService,), dict())

    def __init__(self, cls):
        cls.category = self.name
        cls.__interfaces__ = dict()

    def __str__(self):
        return self._name_

    def __call__(cls, name: str):
        assert len(name) > 0, "invalid service name"
        assert name not in cls.__interfaces__

        def __new__(cls_):
            d_ = dict(cls_.__dict__)
            d_.update({"name": name})
            cls_ = abc.ABCMeta.__new__(abc.ABCMeta, cls_.__name__, (cls.value,), d_)
            cls.__interfaces__[name] = cls_
            return cls_

        return __new__

    @property
    def __interfaces__(self):
        return self.value.__interfaces__

    @staticmethod
    def group(
        category: "ServiceGroup" | typing.Iterable["ServiceGroup"] | str | None,
        services: list[T],
    ) -> list[T]:
        if category is None:

            def filter_(s) -> bool:
                return True

        else:
            if isinstance(category, str):
                categories = (category,)
            elif isinstance(category, ServiceGroup):
                categories = (category.name,)
            else:
                categories = tuple((c if isinstance(c, str) else c.name) for c in category)  # type: ignore[assignment]

            def filter_(s) -> bool:
                s_cat = s.category if isinstance(s.category, str) else s.category.name
                return s_cat in categories

        return list(filter(filter_, services))


@ServiceGroup.host("state")
class IStateService(abc.ABC):
    class states(enum.Enum):
        suspended = 0
        stopped = 1
        started = 2
        starting = 3
        stopping = 4
        suspending = 5
        unknown = -1

    @abc.abstractmethod
    def start(self, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def stop(self, force=False, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def suspend(self, force=False, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def reset(self, force=False, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def wait(self, state, delay=-1, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def status(self):
        ...


@ServiceGroup.host("snap")
class ISnapService(abc.ABC):
    import posixpath

    sep = posixpath.sep
    join = staticmethod(posixpath.join)
    isabs = staticmethod(posixpath.isabs)
    basename = staticmethod(posixpath.basename)
    history = staticmethod(posixpath.dirname)

    @abc.abstractmethod
    def new(self, name, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def delete(self, name, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def restore(self, name, wait=False, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def find(self, name, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def names(self):
        ...

    @abc.abstractmethod
    def current(self):
        ...


@ServiceGroup.host("instance")
class IInstanceService:
    @abc.abstractmethod
    def delete(self, warehouse, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def create(self, warehouse, *, extras: dict | None = None):
        ...


@ServiceGroup.system("shell")
class IShellService(abc.ABC):
    @abc.abstractmethod
    def execute(self, command: bytes, working_dir: bytes | None = None, env: dict | None = None, *, extras: dict | None = None) -> tuple[bytes, bytes, bytes]:
        ...

    @abc.abstractmethod
    def spawn(self, command: bytes, working_dir: bytes | None = None, env: dict | None = None, *, extras: dict | None = None) -> int:
        ...

    @abc.abstractmethod
    def signal(self, procid: int, sig: int, *, extras: dict | None = None):
        ...


@ServiceGroup.system("fs")
class IFileSystemService(abc.ABC):
    @abc.abstractmethod
    def rm(self, path: bytes, *, extras: dict | None = None) -> None:
        ...

    @abc.abstractmethod
    def mkdir(self, path: bytes, *, extras: dict | None = None) -> None:
        ...

    @abc.abstractmethod
    def rmdir(self, path: bytes, *, extras: dict | None = None) -> None:
        ...

    @abc.abstractmethod
    def is_container(self, path: bytes, *, extras: dict | None = None) -> bool:
        ...

    @abc.abstractmethod
    def exist(self, path: bytes, *, extras: dict | None = None) -> bool:
        ...

    @abc.abstractmethod
    def list(self, path: bytes, *, extras: dict | None = None) -> list[bytes]:
        ...

    @abc.abstractmethod
    def stat(self, path: bytes, *, extras: dict | None = None) -> dict:
        ...


@ServiceGroup.system("stream")
class IStreamService(abc.ABC):
    @abc.abstractmethod
    def open_file(self, path: bytes, wiring: int, *, extras: dict | None = None) -> tuple[bytes, int]:
        ...

    @abc.abstractmethod
    def open_process(self, path: bytes, wiring: int, args: list[bytes], working_dir: bytes | None = None, env: dict[bytes, bytes] | None = None, *, extras: dict | None = None) -> tuple[bytes, int, int | None, int | None, int | None]:
        ...

    @abc.abstractmethod
    def read(self, handle: int, nbytes: int, *, extras: dict | None = None) -> bytearray:
        ...

    @abc.abstractmethod
    def readall(self, handle: int, *, extras: dict | None = None) -> bytearray:
        ...

    @abc.abstractmethod
    def readchunk(self, handle: int, *, extras: dict | None = None) -> bytearray:
        ...

    @abc.abstractmethod
    def write(self, handle: int, data: bytes, *, extras: dict | None = None):
        ...

    @abc.abstractmethod
    def close(self, handle: int, *, extras: dict | None = None) -> None:
        ...

    @abc.abstractmethod
    def seek(self, handle: int, pos: int, whence: int, *, extras: dict | None = None) -> int:
        ...

    @abc.abstractmethod
    def sync(self, handle: int, *, extras: dict | None = None) -> None:
        ...

    @abc.abstractmethod
    def flush(self, handle: int, *, extras: dict | None = None) -> bytearray:
        ...

    @abc.abstractmethod
    def truncate(self, handle: int, length: int, *, extras: dict | None = None) -> None:
        ...

    @abc.abstractmethod
    def stat(self, handle: int, fields: int = StatField.ALL.value, *, extras: dict | None = None) -> dict:
        ...

    @abc.abstractmethod
    def terminate(self, handle: int, *, extras: dict | None = None) -> None:
        ...



import re

from myrrh.utils import mshlex

from myrrh.core.services import cfg_init, cfg_get, cfg_del, cfg_set

from myrrh.tools.myrrhc_ext import myrrhc_cmds, cmd
from myrrh.tools.myrrhc.main import getsession


cfg_init("asetcfg", "setcfg --section myrrh.tools.myrrhc.alias", section="myrrh.tools.myrrhc.alias")


def _alias_cmd(name, value):
    args = list()
    nargs = len(re.split(r"{\d*}", value)) - 1
    for n in range(0, nargs):
        args.append(cmd.Argument((f"arg{n}",), required=True))

    if not nargs:
        nkwargs = re.findall(r"\{([a-zA-Z_]\w*)\}", value)
        for arg in nkwargs:
            args.append(cmd.Argument((arg,), required=True))

    @myrrhc_cmds.command(name=name)
    def _(**kwargs):
        f"""
        Alias for "{value}"
        """

        async def _():
            if nargs:
                cmdline = value.format(*kwargs.values())

            elif nkwargs:
                cmdline = f"{value} {mshlex.list2cmdlines(**kwargs)}"

            else:
                cmdline = f"{value}"

            return await getsession().command_call(cmdline)

        return _

    _.params.extend(args)
    _.help = f'Alias for "{value}"'


@myrrhc_cmds.command()
@cmd.argument("alias_name")
@cmd.argument("alias_value", default=None, required=False)
def alias(alias_name, alias_value):
    """
    Create an command alias
    """
    if alias_value is None:
        cfg_del(alias_name, section="myrrh.tools.myrrhc.alias")
        myrrhc_cmds.commands.pop(alias_name)
    else:
        cfg_set(alias_name, alias_value, section="myrrh.tools.myrrhc.alias")

    _reload()


@myrrhc_cmds.command()
def areload():
    """
    Reload all alias
    """
    _reload()


def _reload():
    for name, value in cfg_get(section="myrrh.tools.myrrhc.alias").items():
        _alias_cmd(name, value)


_reload()



import typing
import unittest

from myrrh.factory import Assembly

from myrrh.provider.registry import ProviderRegistry
from myrrh.warehouse.registry import ItemRegistry


class AssemblyTest(unittest.TestCase):
    def test_basic_build_local(self):
        p_local = ProviderRegistry().get("local")()

        f = Assembly.fromProvider(p_local)
        e = f.build()
        f_from_e = f.fromEntity(e, only_predefined=False)

        self.assertEqual(f.supply, f_from_e.supply)
        self.assertEqual(e.cfg.system, f_from_e.get_item("system"))
        self.assertEqual(e.cfg.id, f_from_e.get_item("id"))

    def test_basic_build_local_with_predefined(self):
        f = Assembly(
            supply={"paths": ["**/local"]},
            warehouse=[{"type_": "system", "label": "mylabel"}],
        )
        e = f.build()
        f_from_e = f.fromEntity(e, only_predefined=False)

        self.assertEqual(f.supply, f_from_e.supply)
        self.assertEqual(e.cfg.system, f_from_e.get_item("system"))
        self.assertEqual(e.cfg.system.label, "mylabel")
        self.assertEqual(e.cfg.id, f_from_e.get_item("id"))

    def test_basic_settings_raw(self):
        c = Assembly.fromRaw(
            """{
            "supply" : { "paths" : ["**/local"]},
            "warehouse" : [
                { "type_" : "system"},
                { "type_" : "credentials"},
                { "type_" : "id"},
                { "type_" : "host"},
                { "type_" : "new_kind"}
            ]
        }"""
        )

        self.assertIsInstance(c.supply, ItemRegistry().FactorySupply)
        self.assertIsInstance(c.warehouse[0], ItemRegistry().system)
        self.assertIsInstance(c.warehouse[1], ItemRegistry().credentials)
        self.assertIsInstance(c.warehouse[2], ItemRegistry().id)
        self.assertIsInstance(c.warehouse[3], ItemRegistry().host)
        self.assertIsInstance(c.warehouse[4], ItemRegistry().generic)

    def test_basic_settings_init(self):
        c = Assembly(
            supply={"paths": ["**/local"]},
            warehouse=[
                {"type_": "system"},
                {"type_": "credentials"},
                {"type_": "id"},
                {"type_": "host"},
                {"type_": "new_kind"},
            ],
        )

        self.assertIsInstance(c.supply, ItemRegistry().FactorySupply)
        self.assertIsInstance(c.warehouse[0], ItemRegistry().system)
        self.assertIsInstance(c.warehouse[1], ItemRegistry().credentials)
        self.assertIsInstance(c.warehouse[2], ItemRegistry().id)
        self.assertIsInstance(c.warehouse[3], ItemRegistry().host)
        self.assertIsInstance(c.warehouse[4], ItemRegistry().generic)

    def test_basic_settings_obj(self):
        c = Assembly.fromObj(
            {
                "supply": {"paths": ["**/local"]},
                "warehouse": [
                    {"type_": "system"},
                    {"type_": "credentials"},
                    {"type_": "id"},
                    {"type_": "host"},
                    {"type_": "new_kind"},
                ],
            }
        )
        self.assertIsInstance(c.supply, ItemRegistry().FactorySupply)
        self.assertIsInstance(c.warehouse[0], ItemRegistry().system)
        self.assertIsInstance(c.warehouse[1], ItemRegistry().credentials)
        self.assertIsInstance(c.warehouse[2], ItemRegistry().id)
        self.assertIsInstance(c.warehouse[3], ItemRegistry().host)
        self.assertIsInstance(c.warehouse[4], ItemRegistry().generic)


class ItemsBasicTests(unittest.TestCase):
    def test_basic_create_item(self):
        c = ItemRegistry().warehouse_model_validate({"type_": "system"})
        self.assertEqual(c.type_, "system")
        c = ItemRegistry().warehouse_model_validate({"type_": "new_kind", "key": "value"})
        self.assertEqual(c.type_, "new_kind")
        self.assertEqual(c.key, "value")


class RegisterBasicTests(unittest.TestCase):
    def test_basic_register_item(self):
        from myrrh.warehouse.item import BaseItem

        class PluginItem(BaseItem[typing.Literal["plugin"]]):
            type_: typing.Literal["plugin"]
            val: str

        ItemRegistry().register_warehouse(PluginItem)
        c = ItemRegistry().warehouse_model_validate({"type_": "plugin", "val": "value"})

        self.assertIsInstance(c, PluginItem)
        self.assertEqual(c.type_, "plugin")
        self.assertEqual(c.val, "value")
        self.assertIn("plugin", ItemRegistry().items)

    def test_basic_register_provider_model(self):
        import myrrh.warehouse

        class NewProviderModel(myrrh.warehouse.Settings):
            name: typing.Literal["new_provider"]
            usable: bool = False

        ItemRegistry().register_provider_model(NewProviderModel)

        c = ItemRegistry().provider_model_validate({"name": "new_provider", "usable": True})

        self.assertIsInstance(c, NewProviderModel)
        self.assertEqual(c.name, "new_provider")
        self.assertEqual(c.usable, True)


if __name__ == "__main__":
    unittest.main()



# flake8: noqa:

import bmy

import os
from myrrh.utils.myrrh_test_init import *

setUp(f"bmy.{os.path.basename(__file__)}")

# import test.test_tempfile
# tempfile.py unit tests.
with bmy.select():
    from mlib.py import tempfile
import errno

with bmy.select():
    from mlib.py import io
    from mlib.py import os
import pathlib

with bmy.select():
    from mlib.py import sys
import re
import warnings
import contextlib
import stat
import types
import weakref
from unittest import mock

import unittest

with bmy.select():
    from mlib.py.test import support
    from mlib.py.test.support import os_helper
    from mlib.py.test.support import script_helper
    from mlib.py.test.support import warnings_helper


has_textmode = tempfile._text_openflags != tempfile._bin_openflags
has_spawnl = hasattr(os, "spawnl")

# TEST_FILES may need to be tweaked for systems depending on the maximum
# number of files that can be opened at one time (see ulimit -n)
if sys.platform.startswith("openbsd"):
    TEST_FILES = 48
else:
    TEST_FILES = 100

# This is organized as one test for each chunk of code in tempfile.py,
# in order of their appearance in the file.  Testing which requires
# threads is not done here.


class TestLowLevelInternals(unittest.TestCase):
    def test_infer_return_type_singles(self):
        self.assertIs(str, tempfile._infer_return_type(""))
        self.assertIs(bytes, tempfile._infer_return_type(b""))
        self.assertIs(str, tempfile._infer_return_type(None))

    def test_infer_return_type_multiples(self):
        self.assertIs(str, tempfile._infer_return_type("", ""))
        self.assertIs(bytes, tempfile._infer_return_type(b"", b""))
        with self.assertRaises(TypeError):
            tempfile._infer_return_type("", b"")
        with self.assertRaises(TypeError):
            tempfile._infer_return_type(b"", "")

    def test_infer_return_type_multiples_and_none(self):
        self.assertIs(str, tempfile._infer_return_type(None, ""))
        self.assertIs(str, tempfile._infer_return_type("", None))
        self.assertIs(str, tempfile._infer_return_type(None, None))
        self.assertIs(bytes, tempfile._infer_return_type(b"", None))
        self.assertIs(bytes, tempfile._infer_return_type(None, b""))
        with self.assertRaises(TypeError):
            tempfile._infer_return_type("", None, b"")
        with self.assertRaises(TypeError):
            tempfile._infer_return_type(b"", None, "")

    def test_infer_return_type_pathlib(self):
        self.assertIs(str, tempfile._infer_return_type(pathlib.Path("/")))

    def test_infer_return_type_pathlike(self):
        class Path:
            def __init__(self, path):
                self.path = path

            def __fspath__(self):
                return self.path

        self.assertIs(str, tempfile._infer_return_type(Path("/")))
        self.assertIs(bytes, tempfile._infer_return_type(Path(b"/")))
        self.assertIs(str, tempfile._infer_return_type("", Path("")))
        self.assertIs(bytes, tempfile._infer_return_type(b"", Path(b"")))
        self.assertIs(bytes, tempfile._infer_return_type(None, Path(b"")))
        self.assertIs(str, tempfile._infer_return_type(None, Path("")))

        with self.assertRaises(TypeError):
            tempfile._infer_return_type("", Path(b""))
        with self.assertRaises(TypeError):
            tempfile._infer_return_type(b"", Path(""))


# Common functionality.


class BaseTestCase(unittest.TestCase):
    str_check = re.compile(r"^[a-z0-9_-]{8}$")
    b_check = re.compile(rb"^[a-z0-9_-]{8}$")

    def setUp(self):
        self.enterContext(warnings_helper.check_warnings())
        warnings.filterwarnings("ignore", category=RuntimeWarning, message="mktemp", module=__name__)

    def nameCheck(self, name, dir, pre, suf):
        (ndir, nbase) = os.path.split(name)
        npre = nbase[: len(pre)]
        nsuf = nbase[len(nbase) - len(suf) :]

        if dir is not None:
            self.assertIs(
                type(name),
                str if type(dir) is str or isinstance(dir, os.PathLike) else bytes,
                "unexpected return type",
            )
        if pre is not None:
            self.assertIs(type(name), str if type(pre) is str else bytes, "unexpected return type")
        if suf is not None:
            self.assertIs(type(name), str if type(suf) is str else bytes, "unexpected return type")
        if (dir, pre, suf) == (None, None, None):
            self.assertIs(type(name), str, "default return type must be str")

        # check for equality of the absolute paths!
        self.assertEqual(
            os.path.abspath(ndir),
            os.path.abspath(dir),
            "file %r not in directory %r" % (name, dir),
        )
        self.assertEqual(npre, pre, "file %r does not begin with %r" % (nbase, pre))
        self.assertEqual(nsuf, suf, "file %r does not end with %r" % (nbase, suf))

        nbase = nbase[len(pre) : len(nbase) - len(suf)]
        check = self.str_check if isinstance(nbase, str) else self.b_check
        self.assertTrue(
            check.match(nbase),
            "random characters %r do not match %r" % (nbase, check.pattern),
        )


class TestExports(BaseTestCase):
    def test_exports(self):
        # There are no surprising symbols in the tempfile module
        dict = tempfile.__dict__

        expected = {
            "NamedTemporaryFile": 1,
            "TemporaryFile": 1,
            "mkstemp": 1,
            "mkdtemp": 1,
            "mktemp": 1,
            "TMP_MAX": 1,
            "gettempprefix": 1,
            "gettempprefixb": 1,
            "gettempdir": 1,
            "gettempdirb": 1,
            "tempdir": 1,
            "template": 1,
            "SpooledTemporaryFile": 1,
            "TemporaryDirectory": 1,
        }

        unexp = []
        for key in dict:
            if key[0] != "_" and key not in expected:
                unexp.append(key)
        self.assertTrue(len(unexp) == 0, "unexpected keys: %s" % unexp)


class TestRandomNameSequence(BaseTestCase):
    """Test the internal iterator object _RandomNameSequence."""

    def setUp(self):
        self.r = tempfile._RandomNameSequence()
        super().setUp()

    def test_get_eight_char_str(self):
        # _RandomNameSequence returns a eight-character string
        s = next(self.r)
        self.nameCheck(s, "", "", "")

    def test_many(self):
        # _RandomNameSequence returns no duplicate strings (stochastic)

        dict = {}
        r = self.r
        for i in range(TEST_FILES):
            s = next(r)
            self.nameCheck(s, "", "", "")
            self.assertNotIn(s, dict)
            dict[s] = 1

    def supports_iter(self):
        # _RandomNameSequence supports the iterator protocol

        i = 0
        r = self.r
        for s in r:
            i += 1
            if i == 20:
                break

    @support.requires_fork()
    def test_process_awareness(self):
        # ensure that the random source differs between
        # child and parent.
        read_fd, write_fd = os.pipe()
        pid = None
        try:
            pid = os.fork()
            if not pid:
                # child process
                os.close(read_fd)
                os.write(write_fd, next(self.r).encode("ascii"))
                os.close(write_fd)
                # bypass the normal exit handlers- leave those to
                # the parent.
                os._exit(0)

            # parent process
            parent_value = next(self.r)
            child_value = os.read(read_fd, len(parent_value)).decode("ascii")
        finally:
            if pid:
                support.wait_process(pid, exitcode=0)

            os.close(read_fd)
            os.close(write_fd)
        self.assertNotEqual(child_value, parent_value)


class TestCandidateTempdirList(BaseTestCase):
    """Test the internal function _candidate_tempdir_list."""

    def test_nonempty_list(self):
        # _candidate_tempdir_list returns a nonempty list of strings

        cand = tempfile._candidate_tempdir_list()

        self.assertFalse(len(cand) == 0)
        for c in cand:
            self.assertIsInstance(c, str)

    def test_wanted_dirs(self):
        # _candidate_tempdir_list contains the expected directories

        # Make sure the interesting environment variables are all set.
        with os_helper.EnvironmentVarGuard() as env:
            for envname in "TMPDIR", "TEMP", "TMP":
                dirname = os.getenv(envname)
                if not dirname:
                    env[envname] = os.path.abspath(envname)

            cand = tempfile._candidate_tempdir_list()

            for envname in "TMPDIR", "TEMP", "TMP":
                dirname = os.getenv(envname)
                if not dirname:
                    raise ValueError
                self.assertIn(dirname, cand)

            try:
                dirname = os.getcwd()
            except (AttributeError, OSError):
                dirname = os.curdir

            self.assertIn(dirname, cand)

            # Not practical to try to verify the presence of OS-specific
            # paths in this list.


# We test _get_default_tempdir some more by testing gettempdir.


class TestGetDefaultTempdir(BaseTestCase):
    """Test _get_default_tempdir()."""

    def test_no_files_left_behind(self):
        # use a private empty directory
        with tempfile.TemporaryDirectory() as our_temp_directory:
            # force _get_default_tempdir() to consider our empty directory
            def our_candidate_list():
                return [our_temp_directory]

            with support.swap_attr(tempfile, "_candidate_tempdir_list", our_candidate_list):
                # verify our directory is empty after _get_default_tempdir()
                tempfile._get_default_tempdir()
                self.assertEqual(os.listdir(our_temp_directory), [])

                def raise_OSError(*args, **kwargs):
                    raise OSError()

                with support.swap_attr(os, "open", raise_OSError):
                    # test again with failing os.open()
                    with self.assertRaises(FileNotFoundError):
                        tempfile._get_default_tempdir()
                    self.assertEqual(os.listdir(our_temp_directory), [])

                with support.swap_attr(os, "write", raise_OSError):
                    # test again with failing os.write()
                    with self.assertRaises(FileNotFoundError):
                        tempfile._get_default_tempdir()
                    self.assertEqual(os.listdir(our_temp_directory), [])


class TestGetCandidateNames(BaseTestCase):
    """Test the internal function _get_candidate_names."""

    def test_retval(self):
        # _get_candidate_names returns a _RandomNameSequence object
        obj = tempfile._get_candidate_names()
        self.assertIsInstance(obj, tempfile._RandomNameSequence)

    def test_same_thing(self):
        # _get_candidate_names always returns the same object
        a = tempfile._get_candidate_names()
        b = tempfile._get_candidate_names()

        self.assertTrue(a is b)


@contextlib.contextmanager
def _inside_empty_temp_dir():
    dir = tempfile.mkdtemp()
    try:
        with support.swap_attr(tempfile, "tempdir", dir):
            yield
    finally:
        os_helper.rmtree(dir)


def _mock_candidate_names(*names):
    return support.swap_attr(tempfile, "_get_candidate_names", lambda: iter(names))


class TestBadTempdir:
    @unittest.skipIf(support.is_emscripten, "Emscripten cannot remove write bits.")
    def test_read_only_directory(self):
        with _inside_empty_temp_dir():
            oldmode = mode = os.stat(tempfile.tempdir).st_mode
            mode &= ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
            os.chmod(tempfile.tempdir, mode)
            try:
                if os.access(tempfile.tempdir, os.W_OK):
                    self.skipTest("can't set the directory read-only")
                with self.assertRaises(PermissionError):
                    self.make_temp()
                self.assertEqual(os.listdir(tempfile.tempdir), [])
            finally:
                os.chmod(tempfile.tempdir, oldmode)

    def test_nonexisting_directory(self):
        with _inside_empty_temp_dir():
            tempdir = os.path.join(tempfile.tempdir, "nonexistent")
            with support.swap_attr(tempfile, "tempdir", tempdir):
                with self.assertRaises(FileNotFoundError):
                    self.make_temp()

    def test_non_directory(self):
        with _inside_empty_temp_dir():
            tempdir = os.path.join(tempfile.tempdir, "file")
            open(tempdir, "wb").close()
            with support.swap_attr(tempfile, "tempdir", tempdir):
                with self.assertRaises((NotADirectoryError, FileNotFoundError)):
                    self.make_temp()


class TestMkstempInner(TestBadTempdir, BaseTestCase):
    """Test the internal function _mkstemp_inner."""

    class mkstemped:
        _bflags = tempfile._bin_openflags
        _tflags = tempfile._text_openflags
        _close = os.close
        _unlink = os.unlink

        def __init__(self, dir, pre, suf, bin):
            if bin:
                flags = self._bflags
            else:
                flags = self._tflags

            output_type = tempfile._infer_return_type(dir, pre, suf)
            (self.fd, self.name) = tempfile._mkstemp_inner(dir, pre, suf, flags, output_type)

        def write(self, str):
            os.write(self.fd, str)

        def __del__(self):
            self._close(self.fd)
            self._unlink(self.name)

    def do_create(self, dir=None, pre=None, suf=None, bin=1):
        output_type = tempfile._infer_return_type(dir, pre, suf)
        if dir is None:
            if output_type is str:
                dir = tempfile.gettempdir()
            else:
                dir = tempfile.gettempdirb()
        if pre is None:
            pre = output_type()
        if suf is None:
            suf = output_type()
        file = self.mkstemped(dir, pre, suf, bin)

        self.nameCheck(file.name, dir, pre, suf)
        return file

    def test_basic(self):
        # _mkstemp_inner can create files
        self.do_create().write(b"blat")
        self.do_create(pre="a").write(b"blat")
        self.do_create(suf="b").write(b"blat")
        self.do_create(pre="a", suf="b").write(b"blat")
        self.do_create(pre="aa", suf=".txt").write(b"blat")

    def test_basic_with_bytes_names(self):
        # _mkstemp_inner can create files when given name parts all
        # specified as bytes.
        dir_b = tempfile.gettempdirb()
        self.do_create(dir=dir_b, suf=b"").write(b"blat")
        self.do_create(dir=dir_b, pre=b"a").write(b"blat")
        self.do_create(dir=dir_b, suf=b"b").write(b"blat")
        self.do_create(dir=dir_b, pre=b"a", suf=b"b").write(b"blat")
        self.do_create(dir=dir_b, pre=b"aa", suf=b".txt").write(b"blat")
        # Can't mix str & binary types in the args.
        with self.assertRaises(TypeError):
            self.do_create(dir="", suf=b"").write(b"blat")
        with self.assertRaises(TypeError):
            self.do_create(dir=dir_b, pre="").write(b"blat")
        with self.assertRaises(TypeError):
            self.do_create(dir=dir_b, pre=b"", suf="").write(b"blat")

    def test_basic_many(self):
        # _mkstemp_inner can create many files (stochastic)
        extant = list(range(TEST_FILES))
        for i in extant:
            extant[i] = self.do_create(pre="aa")

    def test_choose_directory(self):
        # _mkstemp_inner can create files in a user-selected directory
        dir = tempfile.mkdtemp()
        try:
            self.do_create(dir=dir).write(b"blat")
            self.do_create(dir=pathlib.Path(dir)).write(b"blat")
        finally:
            support.gc_collect()  # For PyPy or other GCs.
            os.rmdir(dir)

    @os_helper.skip_unless_working_chmod
    def test_file_mode(self):
        # _mkstemp_inner creates files with the proper mode

        file = self.do_create()
        mode = stat.S_IMODE(os.stat(file.name).st_mode)
        expected = 0o600
        if sys.platform == "win32":
            # There's no distinction among 'user', 'group' and 'world';
            # replicate the 'user' bits.
            user = expected >> 6
            expected = user * (1 + 8 + 64)
        self.assertEqual(mode, expected)

    @unittest.skip("myrrh unsupported test")
    @unittest.skipUnless(has_spawnl, "os.spawnl not available")
    @support.requires_subprocess()
    def test_noinherit(self):
        # _mkstemp_inner file handles are not inherited by child processes

        if support.verbose:
            v = "v"
        else:
            v = "q"

        file = self.do_create()
        self.assertEqual(os.get_inheritable(file.fd), False)
        fd = "%d" % file.fd

        try:
            me = __file__
        except NameError:
            me = sys.argv[0]

        # We have to exec something, so that FD_CLOEXEC will take
        # effect.  The core of this test is therefore in
        # tf_inherit_check.py, which see.
        tester = os.path.join(os.path.dirname(os.path.abspath(me)), "tf_inherit_check.py")

        # On Windows a spawn* /path/ with embedded spaces shouldn't be quoted,
        # but an arg with embedded spaces should be decorated with double
        # quotes on each end
        if sys.platform == "win32":
            decorated = '"%s"' % sys.executable
            tester = '"%s"' % tester
        else:
            decorated = sys.executable

        retval = os.spawnl(os.P_WAIT, sys.executable, decorated, tester, v, fd)
        self.assertFalse(retval < 0, "child process caught fatal signal %d" % -retval)
        self.assertFalse(retval > 0, "child process reports failure %d" % retval)

    @unittest.skipUnless(has_textmode, "text mode not available")
    def test_textmode(self):
        # _mkstemp_inner can create files in text mode

        # A text file is truncated at the first Ctrl+Z byte
        f = self.do_create(bin=0)
        f.write(b"blat\x1a")
        f.write(b"extra\n")
        os.lseek(f.fd, 0, os.SEEK_SET)
        self.assertEqual(os.read(f.fd, 20), b"blat")

    def make_temp(self):
        return tempfile._mkstemp_inner(
            tempfile.gettempdir(),
            tempfile.gettempprefix(),
            "",
            tempfile._bin_openflags,
            str,
        )

    def test_collision_with_existing_file(self):
        # _mkstemp_inner tries another name when a file with
        # the chosen name already exists
        with _inside_empty_temp_dir(), _mock_candidate_names("aaa", "aaa", "bbb"):
            (fd1, name1) = self.make_temp()
            os.close(fd1)
            self.assertTrue(name1.endswith("aaa"))

            (fd2, name2) = self.make_temp()
            os.close(fd2)
            self.assertTrue(name2.endswith("bbb"))

    def test_collision_with_existing_directory(self):
        # _mkstemp_inner tries another name when a directory with
        # the chosen name already exists
        with _inside_empty_temp_dir(), _mock_candidate_names("aaa", "aaa", "bbb"):
            dir = tempfile.mkdtemp()
            self.assertTrue(dir.endswith("aaa"))

            (fd, name) = self.make_temp()
            os.close(fd)
            self.assertTrue(name.endswith("bbb"))


class TestGetTempPrefix(BaseTestCase):
    """Test gettempprefix()."""

    def test_sane_template(self):
        # gettempprefix returns a nonempty prefix string
        p = tempfile.gettempprefix()

        self.assertIsInstance(p, str)
        self.assertGreater(len(p), 0)

        pb = tempfile.gettempprefixb()

        self.assertIsInstance(pb, bytes)
        self.assertGreater(len(pb), 0)

    def test_usable_template(self):
        # gettempprefix returns a usable prefix string

        # Create a temp directory, avoiding use of the prefix.
        # Then attempt to create a file whose name is
        # prefix + 'xxxxxx.xxx' in that directory.
        p = tempfile.gettempprefix() + "xxxxxx.xxx"
        d = tempfile.mkdtemp(prefix="")
        try:
            p = os.path.join(d, p)
            fd = os.open(p, os.O_RDWR | os.O_CREAT)
            os.close(fd)
            os.unlink(p)
        finally:
            os.rmdir(d)


class TestGetTempDir(BaseTestCase):
    """Test gettempdir()."""

    def test_directory_exists(self):
        # gettempdir returns a directory which exists

        for d in (tempfile.gettempdir(), tempfile.gettempdirb()):
            self.assertTrue(os.path.isabs(d) or d == os.curdir, "%r is not an absolute path" % d)
            self.assertTrue(os.path.isdir(d), "%r is not a directory" % d)

    def test_directory_writable(self):
        # gettempdir returns a directory writable by the user

        # sneaky: just instantiate a NamedTemporaryFile, which
        # defaults to writing into the directory returned by
        # gettempdir.
        with tempfile.NamedTemporaryFile() as file:
            file.write(b"blat")

    def test_same_thing(self):
        # gettempdir always returns the same object
        a = tempfile.gettempdir()
        b = tempfile.gettempdir()
        c = tempfile.gettempdirb()

        self.assertTrue(a is b)
        self.assertNotEqual(type(a), type(c))
        self.assertEqual(a, os.fsdecode(c))

    def test_case_sensitive(self):
        # gettempdir should not flatten its case
        # even on a case-insensitive file system
        case_sensitive_tempdir = tempfile.mkdtemp("-Temp")
        _tempdir, tempfile.tempdir = tempfile.tempdir, None
        try:
            with os_helper.EnvironmentVarGuard() as env:
                # Fake the first env var which is checked as a candidate
                env["TMPDIR"] = case_sensitive_tempdir
                self.assertEqual(tempfile.gettempdir(), case_sensitive_tempdir)
        finally:
            tempfile.tempdir = _tempdir
            os_helper.rmdir(case_sensitive_tempdir)


class TestMkstemp(BaseTestCase):
    """Test mkstemp()."""

    def do_create(self, dir=None, pre=None, suf=None):
        output_type = tempfile._infer_return_type(dir, pre, suf)
        if dir is None:
            if output_type is str:
                dir = tempfile.gettempdir()
            else:
                dir = tempfile.gettempdirb()
        if pre is None:
            pre = output_type()
        if suf is None:
            suf = output_type()
        (fd, name) = tempfile.mkstemp(dir=dir, prefix=pre, suffix=suf)
        (ndir, nbase) = os.path.split(name)
        adir = os.path.abspath(dir)
        self.assertEqual(adir, ndir, "Directory '%s' incorrectly returned as '%s'" % (adir, ndir))

        try:
            self.nameCheck(name, dir, pre, suf)
        finally:
            os.close(fd)
            os.unlink(name)

    def test_basic(self):
        # mkstemp can create files
        self.do_create()
        self.do_create(pre="a")
        self.do_create(suf="b")
        self.do_create(pre="a", suf="b")
        self.do_create(pre="aa", suf=".txt")
        self.do_create(dir=".")

    def test_basic_with_bytes_names(self):
        # mkstemp can create files when given name parts all
        # specified as bytes.
        d = tempfile.gettempdirb()
        self.do_create(dir=d, suf=b"")
        self.do_create(dir=d, pre=b"a")
        self.do_create(dir=d, suf=b"b")
        self.do_create(dir=d, pre=b"a", suf=b"b")
        self.do_create(dir=d, pre=b"aa", suf=b".txt")
        self.do_create(dir=b".")
        with self.assertRaises(TypeError):
            self.do_create(dir=".", pre=b"aa", suf=b".txt")
        with self.assertRaises(TypeError):
            self.do_create(dir=b".", pre="aa", suf=b".txt")
        with self.assertRaises(TypeError):
            self.do_create(dir=b".", pre=b"aa", suf=".txt")

    def test_choose_directory(self):
        # mkstemp can create directories in a user-selected directory
        dir = tempfile.mkdtemp()
        try:
            self.do_create(dir=dir)
            self.do_create(dir=pathlib.Path(dir))
        finally:
            os.rmdir(dir)

    def test_for_tempdir_is_bytes_issue40701_api_warts(self):
        orig_tempdir = tempfile.tempdir
        self.assertIsInstance(tempfile.tempdir, (str, type(None)))
        try:
            fd, path = tempfile.mkstemp()
            os.close(fd)
            os.unlink(path)
            self.assertIsInstance(path, str)
            tempfile.tempdir = tempfile.gettempdirb()
            self.assertIsInstance(tempfile.tempdir, bytes)
            self.assertIsInstance(tempfile.gettempdir(), str)
            self.assertIsInstance(tempfile.gettempdirb(), bytes)
            fd, path = tempfile.mkstemp()
            os.close(fd)
            os.unlink(path)
            self.assertIsInstance(path, bytes)
            fd, path = tempfile.mkstemp(suffix=".txt")
            os.close(fd)
            os.unlink(path)
            self.assertIsInstance(path, str)
            fd, path = tempfile.mkstemp(prefix="test-temp-")
            os.close(fd)
            os.unlink(path)
            self.assertIsInstance(path, str)
            fd, path = tempfile.mkstemp(dir=tempfile.gettempdir())
            os.close(fd)
            os.unlink(path)
            self.assertIsInstance(path, str)
        finally:
            tempfile.tempdir = orig_tempdir


class TestMkdtemp(TestBadTempdir, BaseTestCase):
    """Test mkdtemp()."""

    def make_temp(self):
        return tempfile.mkdtemp()

    def do_create(self, dir=None, pre=None, suf=None):
        output_type = tempfile._infer_return_type(dir, pre, suf)
        if dir is None:
            if output_type is str:
                dir = tempfile.gettempdir()
            else:
                dir = tempfile.gettempdirb()
        if pre is None:
            pre = output_type()
        if suf is None:
            suf = output_type()
        name = tempfile.mkdtemp(dir=dir, prefix=pre, suffix=suf)

        try:
            self.nameCheck(name, dir, pre, suf)
            return name
        except:
            os.rmdir(name)
            raise

    def test_basic(self):
        # mkdtemp can create directories
        os.rmdir(self.do_create())
        os.rmdir(self.do_create(pre="a"))
        os.rmdir(self.do_create(suf="b"))
        os.rmdir(self.do_create(pre="a", suf="b"))
        os.rmdir(self.do_create(pre="aa", suf=".txt"))

    def test_basic_with_bytes_names(self):
        # mkdtemp can create directories when given all binary parts
        d = tempfile.gettempdirb()
        os.rmdir(self.do_create(dir=d))
        os.rmdir(self.do_create(dir=d, pre=b"a"))
        os.rmdir(self.do_create(dir=d, suf=b"b"))
        os.rmdir(self.do_create(dir=d, pre=b"a", suf=b"b"))
        os.rmdir(self.do_create(dir=d, pre=b"aa", suf=b".txt"))
        with self.assertRaises(TypeError):
            os.rmdir(self.do_create(dir=d, pre="aa", suf=b".txt"))
        with self.assertRaises(TypeError):
            os.rmdir(self.do_create(dir=d, pre=b"aa", suf=".txt"))
        with self.assertRaises(TypeError):
            os.rmdir(self.do_create(dir="", pre=b"aa", suf=b".txt"))

    def test_basic_many(self):
        # mkdtemp can create many directories (stochastic)
        extant = list(range(TEST_FILES))
        try:
            for i in extant:
                extant[i] = self.do_create(pre="aa")
        finally:
            for i in extant:
                if isinstance(i, str):
                    os.rmdir(i)

    def test_choose_directory(self):
        # mkdtemp can create directories in a user-selected directory
        dir = tempfile.mkdtemp()
        try:
            os.rmdir(self.do_create(dir=dir))
            os.rmdir(self.do_create(dir=pathlib.Path(dir)))
        finally:
            os.rmdir(dir)

    @os_helper.skip_unless_working_chmod
    def test_mode(self):
        # mkdtemp creates directories with the proper mode

        dir = self.do_create()
        try:
            mode = stat.S_IMODE(os.stat(dir).st_mode)
            mode &= 0o777  # Mask off sticky bits inherited from /tmp
            expected = 0o700
            if sys.platform == "win32":
                # There's no distinction among 'user', 'group' and 'world';
                # replicate the 'user' bits.
                user = expected >> 6
                expected = user * (1 + 8 + 64)
            self.assertEqual(mode, expected)
        finally:
            os.rmdir(dir)

    def test_collision_with_existing_file(self):
        # mkdtemp tries another name when a file with
        # the chosen name already exists
        with _inside_empty_temp_dir(), _mock_candidate_names("aaa", "aaa", "bbb"):
            file = tempfile.NamedTemporaryFile(delete=False)
            file.close()
            self.assertTrue(file.name.endswith("aaa"))
            dir = tempfile.mkdtemp()
            self.assertTrue(dir.endswith("bbb"))

    def test_collision_with_existing_directory(self):
        # mkdtemp tries another name when a directory with
        # the chosen name already exists
        with _inside_empty_temp_dir(), _mock_candidate_names("aaa", "aaa", "bbb"):
            dir1 = tempfile.mkdtemp()
            self.assertTrue(dir1.endswith("aaa"))
            dir2 = tempfile.mkdtemp()
            self.assertTrue(dir2.endswith("bbb"))

    def test_for_tempdir_is_bytes_issue40701_api_warts(self):
        orig_tempdir = tempfile.tempdir
        self.assertIsInstance(tempfile.tempdir, (str, type(None)))
        try:
            path = tempfile.mkdtemp()
            os.rmdir(path)
            self.assertIsInstance(path, str)
            tempfile.tempdir = tempfile.gettempdirb()
            self.assertIsInstance(tempfile.tempdir, bytes)
            self.assertIsInstance(tempfile.gettempdir(), str)
            self.assertIsInstance(tempfile.gettempdirb(), bytes)
            path = tempfile.mkdtemp()
            os.rmdir(path)
            self.assertIsInstance(path, bytes)
            path = tempfile.mkdtemp(suffix="-dir")
            os.rmdir(path)
            self.assertIsInstance(path, str)
            path = tempfile.mkdtemp(prefix="test-mkdtemp-")
            os.rmdir(path)
            self.assertIsInstance(path, str)
            path = tempfile.mkdtemp(dir=tempfile.gettempdir())
            os.rmdir(path)
            self.assertIsInstance(path, str)
        finally:
            tempfile.tempdir = orig_tempdir


class TestMktemp(BaseTestCase):
    """Test mktemp()."""

    # For safety, all use of mktemp must occur in a private directory.
    # We must also suppress the RuntimeWarning it generates.
    def setUp(self):
        self.dir = tempfile.mkdtemp()
        super().setUp()

    def tearDown(self):
        if self.dir:
            os.rmdir(self.dir)
            self.dir = None
        super().tearDown()

    class mktemped:
        _unlink = os.unlink
        _bflags = tempfile._bin_openflags

        def __init__(self, dir, pre, suf):
            self.name = tempfile.mktemp(dir=dir, prefix=pre, suffix=suf)
            # Create the file.  This will raise an exception if it's
            # mysteriously appeared in the meanwhile.
            os.close(os.open(self.name, self._bflags, 0o600))

        def __del__(self):
            self._unlink(self.name)

    def do_create(self, pre="", suf=""):
        file = self.mktemped(self.dir, pre, suf)

        self.nameCheck(file.name, self.dir, pre, suf)
        return file

    def test_basic(self):
        # mktemp can choose usable file names
        self.do_create()
        self.do_create(pre="a")
        self.do_create(suf="b")
        self.do_create(pre="a", suf="b")
        self.do_create(pre="aa", suf=".txt")
        support.gc_collect()

    def test_many(self):
        # mktemp can choose many usable file names (stochastic)
        extant = list(range(TEST_FILES))
        for i in extant:
            extant[i] = self.do_create(pre="aa")
        del extant
        support.gc_collect()  # For PyPy or other GCs.


##     def test_warning(self):
##         # mktemp issues a warning when used
##         warnings.filterwarnings("error",
##                                 category=RuntimeWarning,
##                                 message="mktemp")
##         self.assertRaises(RuntimeWarning,
##                           tempfile.mktemp, dir=self.dir)


# We test _TemporaryFileWrapper by testing NamedTemporaryFile.


class TestNamedTemporaryFile(BaseTestCase):
    """Test NamedTemporaryFile()."""

    def do_create(self, dir=None, pre="", suf="", delete=True):
        if dir is None:
            dir = tempfile.gettempdir()
        file = tempfile.NamedTemporaryFile(dir=dir, prefix=pre, suffix=suf, delete=delete)

        self.nameCheck(file.name, dir, pre, suf)
        return file

    def test_basic(self):
        # NamedTemporaryFile can create files
        self.do_create()
        self.do_create(pre="a")
        self.do_create(suf="b")
        self.do_create(pre="a", suf="b")
        self.do_create(pre="aa", suf=".txt")

    def test_method_lookup(self):
        # Issue #18879: Looking up a temporary file method should keep it
        # alive long enough.
        f = self.do_create()
        wr = weakref.ref(f)
        write = f.write
        write2 = f.write
        del f
        write(b"foo")
        del write
        write2(b"bar")
        del write2
        if support.check_impl_detail(cpython=True):
            # No reference cycle was created.
            self.assertIsNone(wr())

    def test_iter(self):
        # Issue #23700: getting iterator from a temporary file should keep
        # it alive as long as it's being iterated over
        lines = [b"spam\n", b"eggs\n", b"beans\n"]

        def make_file():
            f = tempfile.NamedTemporaryFile(mode="w+b")
            f.write(b"".join(lines))
            f.seek(0)
            return f

        for i, l in enumerate(make_file()):
            self.assertEqual(l, lines[i])
        self.assertEqual(i, len(lines) - 1)

    def test_creates_named(self):
        # NamedTemporaryFile creates files with names
        f = tempfile.NamedTemporaryFile()
        self.assertTrue(os.path.exists(f.name), "NamedTemporaryFile %s does not exist" % f.name)

    def test_del_on_close(self):
        # A NamedTemporaryFile is deleted when closed
        dir = tempfile.mkdtemp()
        try:
            with tempfile.NamedTemporaryFile(dir=dir) as f:
                f.write(b"blat")
            self.assertEqual(os.listdir(dir), [])
            self.assertFalse(
                os.path.exists(f.name),
                "NamedTemporaryFile %s exists after close" % f.name,
            )
        finally:
            os.rmdir(dir)

    def test_dis_del_on_close(self):
        # Tests that delete-on-close can be disabled
        dir = tempfile.mkdtemp()
        tmp = None
        try:
            f = tempfile.NamedTemporaryFile(dir=dir, delete=False)
            tmp = f.name
            f.write(b"blat")
            f.close()
            self.assertTrue(
                os.path.exists(f.name),
                "NamedTemporaryFile %s missing after close" % f.name,
            )
        finally:
            if tmp is not None:
                os.unlink(tmp)
            os.rmdir(dir)

    def test_multiple_close(self):
        # A NamedTemporaryFile can be closed many times without error
        f = tempfile.NamedTemporaryFile()
        f.write(b"abc\n")
        f.close()
        f.close()
        f.close()

    def test_context_manager(self):
        # A NamedTemporaryFile can be used as a context manager
        with tempfile.NamedTemporaryFile() as f:
            self.assertTrue(os.path.exists(f.name))
        self.assertFalse(os.path.exists(f.name))

        def use_closed():
            with f:
                pass

        self.assertRaises(ValueError, use_closed)

    def test_bad_mode(self):
        dir = tempfile.mkdtemp()
        self.addCleanup(os_helper.rmtree, dir)
        with self.assertRaises(ValueError):
            tempfile.NamedTemporaryFile(mode="wr", dir=dir)
        with self.assertRaises(TypeError):
            tempfile.NamedTemporaryFile(mode=2, dir=dir)
        self.assertEqual(os.listdir(dir), [])

    def test_bad_encoding(self):
        dir = tempfile.mkdtemp()
        self.addCleanup(os_helper.rmtree, dir)
        with self.assertRaises(LookupError):
            tempfile.NamedTemporaryFile("w", encoding="bad-encoding", dir=dir)
        self.assertEqual(os.listdir(dir), [])

    def test_unexpected_error(self):
        dir = tempfile.mkdtemp()
        self.addCleanup(os_helper.rmtree, dir)
        with mock.patch.object(tempfile, "_TemporaryFileWrapper") as mock_ntf, mock.patch.object(io, "open", mock.mock_open()) as mock_open:
            mock_ntf.side_effect = KeyboardInterrupt()
            with self.assertRaises(KeyboardInterrupt):
                tempfile.NamedTemporaryFile(dir=dir)
        mock_open().close.assert_called()
        self.assertEqual(os.listdir(dir), [])

    # How to test the mode and bufsize parameters?


class TestSpooledTemporaryFile(BaseTestCase):
    """Test SpooledTemporaryFile()."""

    def do_create(self, max_size=0, dir=None, pre="", suf=""):
        if dir is None:
            dir = tempfile.gettempdir()
        file = tempfile.SpooledTemporaryFile(max_size=max_size, dir=dir, prefix=pre, suffix=suf)

        return file

    def test_basic(self):
        # SpooledTemporaryFile can create files
        f = self.do_create()
        self.assertFalse(f._rolled)
        f = self.do_create(max_size=100, pre="a", suf=".txt")
        self.assertFalse(f._rolled)

    def test_is_iobase(self):
        # SpooledTemporaryFile should implement io.IOBase
        self.assertIsInstance(self.do_create(), io.IOBase)

    def test_iobase_interface(self):
        # SpooledTemporaryFile should implement the io.IOBase interface.
        # Ensure it has all the required methods and properties.
        iobase_attrs = {
            # From IOBase
            "fileno",
            "seek",
            "truncate",
            "close",
            "closed",
            "__enter__",
            "__exit__",
            "flush",
            "isatty",
            "__iter__",
            "__next__",
            "readable",
            "readline",
            "readlines",
            "seekable",
            "tell",
            "writable",
            "writelines",
            # From BufferedIOBase (binary mode) and TextIOBase (text mode)
            "detach",
            "read",
            "read1",
            "write",
            "readinto",
            "readinto1",
            "encoding",
            "errors",
            "newlines",
        }
        spooledtempfile_attrs = set(dir(tempfile.SpooledTemporaryFile))
        missing_attrs = iobase_attrs - spooledtempfile_attrs
        self.assertFalse(
            missing_attrs,
            "SpooledTemporaryFile missing attributes from IOBase/BufferedIOBase/TextIOBase",
        )

    def test_del_on_close(self):
        # A SpooledTemporaryFile is deleted when closed
        dir = tempfile.mkdtemp()
        try:
            f = tempfile.SpooledTemporaryFile(max_size=10, dir=dir)
            self.assertFalse(f._rolled)
            f.write(b"blat " * 5)
            self.assertTrue(f._rolled)
            filename = f.name
            f.close()
            self.assertEqual(os.listdir(dir), [])
            if not isinstance(filename, int):
                self.assertFalse(
                    os.path.exists(filename),
                    "SpooledTemporaryFile %s exists after close" % filename,
                )
        finally:
            os.rmdir(dir)

    def test_del_unrolled_file(self):
        # The unrolled SpooledTemporaryFile should raise a ResourceWarning
        # when deleted since the file was not explicitly closed.
        f = self.do_create(max_size=10)
        f.write(b"foo")
        self.assertEqual(f.name, None)  # Unrolled so no filename/fd
        with self.assertWarns(ResourceWarning):
            f.__del__()

    @unittest.skipIf(support.is_emscripten, "Emscripten cannot fstat renamed files.")
    def test_del_rolled_file(self):
        # The rolled file should be deleted when the SpooledTemporaryFile
        # object is deleted. This should raise a ResourceWarning since the file
        # was not explicitly closed.
        f = self.do_create(max_size=2)
        f.write(b"foo")
        name = f.name  # This is a fd on posix+cygwin, a filename everywhere else
        self.assertTrue(os.path.exists(name))
        with self.assertWarns(ResourceWarning):
            f.__del__()
        self.assertFalse(
            os.path.exists(name),
            "Rolled SpooledTemporaryFile (name=%s) exists after delete" % name,
        )

    def test_rewrite_small(self):
        # A SpooledTemporaryFile can be written to multiple within the max_size
        f = self.do_create(max_size=30)
        self.assertFalse(f._rolled)
        for i in range(5):
            f.seek(0, 0)
            f.write(b"x" * 20)
        self.assertFalse(f._rolled)

    def test_write_sequential(self):
        # A SpooledTemporaryFile should hold exactly max_size bytes, and roll
        # over afterward
        f = self.do_create(max_size=30)
        self.assertFalse(f._rolled)
        f.write(b"x" * 20)
        self.assertFalse(f._rolled)
        f.write(b"x" * 10)
        self.assertFalse(f._rolled)
        f.write(b"x")
        self.assertTrue(f._rolled)

    def test_writelines(self):
        # Verify writelines with a SpooledTemporaryFile
        f = self.do_create()
        f.writelines((b"x", b"y", b"z"))
        pos = f.seek(0)
        self.assertEqual(pos, 0)
        buf = f.read()
        self.assertEqual(buf, b"xyz")

    def test_writelines_sequential(self):
        # A SpooledTemporaryFile should hold exactly max_size bytes, and roll
        # over afterward
        f = self.do_create(max_size=35)
        f.writelines((b"x" * 20, b"x" * 10, b"x" * 5))
        self.assertFalse(f._rolled)
        f.write(b"x")
        self.assertTrue(f._rolled)

    def test_sparse(self):
        # A SpooledTemporaryFile that is written late in the file will extend
        # when that occurs
        f = self.do_create(max_size=30)
        self.assertFalse(f._rolled)
        pos = f.seek(100, 0)
        self.assertEqual(pos, 100)
        self.assertFalse(f._rolled)
        f.write(b"x")
        self.assertTrue(f._rolled)

    def test_fileno(self):
        # A SpooledTemporaryFile should roll over to a real file on fileno()
        f = self.do_create(max_size=30)
        self.assertFalse(f._rolled)
        self.assertTrue(f.fileno() > 0)
        self.assertTrue(f._rolled)

    def test_multiple_close_before_rollover(self):
        # A SpooledTemporaryFile can be closed many times without error
        f = tempfile.SpooledTemporaryFile()
        f.write(b"abc\n")
        self.assertFalse(f._rolled)
        f.close()
        f.close()
        f.close()

    def test_multiple_close_after_rollover(self):
        # A SpooledTemporaryFile can be closed many times without error
        f = tempfile.SpooledTemporaryFile(max_size=1)
        f.write(b"abc\n")
        self.assertTrue(f._rolled)
        f.close()
        f.close()
        f.close()

    def test_bound_methods(self):
        # It should be OK to steal a bound method from a SpooledTemporaryFile
        # and use it independently; when the file rolls over, those bound
        # methods should continue to function
        f = self.do_create(max_size=30)
        read = f.read
        write = f.write
        seek = f.seek

        write(b"a" * 35)
        write(b"b" * 35)
        seek(0, 0)
        self.assertEqual(read(70), b"a" * 35 + b"b" * 35)

    def test_properties(self):
        f = tempfile.SpooledTemporaryFile(max_size=10)
        f.write(b"x" * 10)
        self.assertFalse(f._rolled)
        self.assertEqual(f.mode, "w+b")
        self.assertIsNone(f.name)
        with self.assertRaises(AttributeError):
            f.newlines
        with self.assertRaises(AttributeError):
            f.encoding
        with self.assertRaises(AttributeError):
            f.errors

        f.write(b"x")
        self.assertTrue(f._rolled)
        self.assertEqual(f.mode, "rb+")
        self.assertIsNotNone(f.name)
        with self.assertRaises(AttributeError):
            f.newlines
        with self.assertRaises(AttributeError):
            f.encoding
        with self.assertRaises(AttributeError):
            f.errors

    def test_text_mode(self):
        # Creating a SpooledTemporaryFile with a text mode should produce
        # a file object reading and writing (Unicode) text strings.
        f = tempfile.SpooledTemporaryFile(mode="w+", max_size=10, encoding="utf-8")
        f.write("abc\n")
        f.seek(0)
        self.assertEqual(f.read(), "abc\n")
        f.write("def\n")
        f.seek(0)
        self.assertEqual(f.read(), "abc\ndef\n")
        self.assertFalse(f._rolled)
        self.assertEqual(f.mode, "w+")
        self.assertIsNone(f.name)
        self.assertEqual(f.newlines, os.linesep)
        self.assertEqual(f.encoding, "utf-8")
        self.assertEqual(f.errors, "strict")

        f.write("xyzzy\n")
        f.seek(0)
        self.assertEqual(f.read(), "abc\ndef\nxyzzy\n")
        # Check that Ctrl+Z doesn't truncate the file
        f.write("foo\x1abar\n")
        f.seek(0)
        self.assertEqual(f.read(), "abc\ndef\nxyzzy\nfoo\x1abar\n")
        self.assertTrue(f._rolled)
        self.assertEqual(f.mode, "w+")
        self.assertIsNotNone(f.name)
        self.assertEqual(f.newlines, os.linesep)
        self.assertEqual(f.encoding, "utf-8")
        self.assertEqual(f.errors, "strict")

    def test_text_newline_and_encoding(self):
        f = tempfile.SpooledTemporaryFile(mode="w+", max_size=10, newline="", encoding="utf-8", errors="ignore")
        f.write("\u039B\r\n")
        f.seek(0)
        self.assertEqual(f.read(), "\u039B\r\n")
        self.assertFalse(f._rolled)
        self.assertEqual(f.mode, "w+")
        self.assertIsNone(f.name)
        self.assertIsNotNone(f.newlines)
        self.assertEqual(f.encoding, "utf-8")
        self.assertEqual(f.errors, "ignore")

        f.write("\u039C" * 10 + "\r\n")
        f.write("\u039D" * 20)
        f.seek(0)
        self.assertEqual(f.read(), "\u039B\r\n" + ("\u039C" * 10) + "\r\n" + ("\u039D" * 20))
        self.assertTrue(f._rolled)
        self.assertEqual(f.mode, "w+")
        self.assertIsNotNone(f.name)
        self.assertIsNotNone(f.newlines)
        self.assertEqual(f.encoding, "utf-8")
        self.assertEqual(f.errors, "ignore")

    def test_context_manager_before_rollover(self):
        # A SpooledTemporaryFile can be used as a context manager
        with tempfile.SpooledTemporaryFile(max_size=1) as f:
            self.assertFalse(f._rolled)
            self.assertFalse(f.closed)
        self.assertTrue(f.closed)

        def use_closed():
            with f:
                pass

        self.assertRaises(ValueError, use_closed)

    def test_context_manager_during_rollover(self):
        # A SpooledTemporaryFile can be used as a context manager
        with tempfile.SpooledTemporaryFile(max_size=1) as f:
            self.assertFalse(f._rolled)
            f.write(b"abc\n")
            f.flush()
            self.assertTrue(f._rolled)
            self.assertFalse(f.closed)
        self.assertTrue(f.closed)

        def use_closed():
            with f:
                pass

        self.assertRaises(ValueError, use_closed)

    def test_context_manager_after_rollover(self):
        # A SpooledTemporaryFile can be used as a context manager
        f = tempfile.SpooledTemporaryFile(max_size=1)
        f.write(b"abc\n")
        f.flush()
        self.assertTrue(f._rolled)
        with f:
            self.assertFalse(f.closed)
        self.assertTrue(f.closed)

        def use_closed():
            with f:
                pass

        self.assertRaises(ValueError, use_closed)

    @unittest.skipIf(support.is_emscripten, "Emscripten cannot fstat renamed files.")
    def test_truncate_with_size_parameter(self):
        # A SpooledTemporaryFile can be truncated to zero size
        f = tempfile.SpooledTemporaryFile(max_size=10)
        f.write(b"abcdefg\n")
        f.seek(0)
        f.truncate()
        self.assertFalse(f._rolled)
        self.assertEqual(f._file.getvalue(), b"")
        # A SpooledTemporaryFile can be truncated to a specific size
        f = tempfile.SpooledTemporaryFile(max_size=10)
        f.write(b"abcdefg\n")
        f.truncate(4)
        self.assertFalse(f._rolled)
        self.assertEqual(f._file.getvalue(), b"abcd")
        # A SpooledTemporaryFile rolls over if truncated to large size
        f = tempfile.SpooledTemporaryFile(max_size=10)
        f.write(b"abcdefg\n")
        f.truncate(20)
        self.assertTrue(f._rolled)
        self.assertEqual(os.fstat(f.fileno()).st_size, 20)

    def test_class_getitem(self):
        self.assertIsInstance(tempfile.SpooledTemporaryFile[bytes], types.GenericAlias)


if tempfile.NamedTemporaryFile is not tempfile.TemporaryFile:

    class TestTemporaryFile(BaseTestCase):
        """Test TemporaryFile()."""

        def test_basic(self):
            # TemporaryFile can create files
            # No point in testing the name params - the file has no name.
            tempfile.TemporaryFile()

        def test_has_no_name(self):
            # TemporaryFile creates files with no names (on this system)
            dir = tempfile.mkdtemp()
            f = tempfile.TemporaryFile(dir=dir)
            f.write(b"blat")

            # Sneaky: because this file has no name, it should not prevent
            # us from removing the directory it was created in.
            try:
                os.rmdir(dir)
            except:
                # cleanup
                f.close()
                os.rmdir(dir)
                raise

        def test_multiple_close(self):
            # A TemporaryFile can be closed many times without error
            f = tempfile.TemporaryFile()
            f.write(b"abc\n")
            f.close()
            f.close()
            f.close()

        # How to test the mode and bufsize parameters?
        def test_mode_and_encoding(self):
            def roundtrip(input, *args, **kwargs):
                with tempfile.TemporaryFile(*args, **kwargs) as fileobj:
                    fileobj.write(input)
                    fileobj.seek(0)
                    self.assertEqual(input, fileobj.read())

            roundtrip(b"1234", "w+b")
            roundtrip("abdc\n", "w+")
            roundtrip("\u039B", "w+", encoding="utf-16")
            roundtrip("foo\r\n", "w+", newline="")

        def test_bad_mode(self):
            dir = tempfile.mkdtemp()
            self.addCleanup(os_helper.rmtree, dir)
            with self.assertRaises(ValueError):
                tempfile.TemporaryFile(mode="wr", dir=dir)
            with self.assertRaises(TypeError):
                tempfile.TemporaryFile(mode=2, dir=dir)
            self.assertEqual(os.listdir(dir), [])

        def test_bad_encoding(self):
            dir = tempfile.mkdtemp()
            self.addCleanup(os_helper.rmtree, dir)
            with self.assertRaises(LookupError):
                tempfile.TemporaryFile("w", encoding="bad-encoding", dir=dir)
            self.assertEqual(os.listdir(dir), [])

        def test_unexpected_error(self):
            dir = tempfile.mkdtemp()
            self.addCleanup(os_helper.rmtree, dir)
            with mock.patch("tempfile._O_TMPFILE_WORKS", False), mock.patch("os.unlink") as mock_unlink, mock.patch("os.open") as mock_open, mock.patch("os.close") as mock_close:
                mock_unlink.side_effect = KeyboardInterrupt()
                with self.assertRaises(KeyboardInterrupt):
                    tempfile.TemporaryFile(dir=dir)
            mock_close.assert_called()
            self.assertEqual(os.listdir(dir), [])


# Helper for test_del_on_shutdown
class NulledModules:
    def __init__(self, *modules):
        self.refs = [mod.__dict__ for mod in modules]
        self.contents = [ref.copy() for ref in self.refs]

    def __enter__(self):
        for d in self.refs:
            for key in d:
                d[key] = None

    def __exit__(self, *exc_info):
        for d, c in zip(self.refs, self.contents):
            d.clear()
            d.update(c)


class TestTemporaryDirectory(BaseTestCase):
    """Test TemporaryDirectory()."""

    def do_create(
        self,
        dir=None,
        pre="",
        suf="",
        recurse=1,
        dirs=1,
        files=1,
        ignore_cleanup_errors=False,
    ):
        if dir is None:
            dir = tempfile.gettempdir()
        tmp = tempfile.TemporaryDirectory(dir=dir, prefix=pre, suffix=suf, ignore_cleanup_errors=ignore_cleanup_errors)
        self.nameCheck(tmp.name, dir, pre, suf)
        self.do_create2(tmp.name, recurse, dirs, files)
        return tmp

    def do_create2(self, path, recurse=1, dirs=1, files=1):
        # Create subdirectories and some files
        if recurse:
            for i in range(dirs):
                name = os.path.join(path, "dir%d" % i)
                os.mkdir(name)
                self.do_create2(name, recurse - 1, dirs, files)
        for i in range(files):
            with open(os.path.join(path, "test%d.txt" % i), "wb") as f:
                f.write(b"Hello world!")

    def test_mkdtemp_failure(self):
        # Check no additional exception if mkdtemp fails
        # Previously would raise AttributeError instead
        # (noted as part of Issue #10188)
        with tempfile.TemporaryDirectory() as nonexistent:
            pass
        with self.assertRaises(FileNotFoundError) as cm:
            tempfile.TemporaryDirectory(dir=nonexistent)
        self.assertEqual(cm.exception.errno, errno.ENOENT)

    def test_explicit_cleanup(self):
        # A TemporaryDirectory is deleted when cleaned up
        dir = tempfile.mkdtemp()
        try:
            d = self.do_create(dir=dir)
            self.assertTrue(os.path.exists(d.name), "TemporaryDirectory %s does not exist" % d.name)
            d.cleanup()
            self.assertFalse(
                os.path.exists(d.name),
                "TemporaryDirectory %s exists after cleanup" % d.name,
            )
        finally:
            os.rmdir(dir)

    def test_explict_cleanup_ignore_errors(self):
        """Test that cleanup doesn't return an error when ignoring them."""
        with tempfile.TemporaryDirectory() as working_dir:
            temp_dir = self.do_create(dir=working_dir, ignore_cleanup_errors=True)
            temp_path = pathlib.Path(temp_dir.name)
            self.assertTrue(temp_path.exists(), f"TemporaryDirectory {temp_path!s} does not exist")
            with open(temp_path / "a_file.txt", "w+t") as open_file:
                open_file.write("Hello world!\n")
                temp_dir.cleanup()
            self.assertEqual(
                len(list(temp_path.glob("*"))),
                int(sys.platform.startswith("win")),
                "Unexpected number of files in " f"TemporaryDirectory {temp_path!s}",
            )
            self.assertEqual(
                temp_path.exists(),
                sys.platform.startswith("win"),
                f"TemporaryDirectory {temp_path!s} existence state unexpected",
            )
            temp_dir.cleanup()
            self.assertFalse(
                temp_path.exists(),
                f"TemporaryDirectory {temp_path!s} exists after cleanup",
            )

    @os_helper.skip_unless_symlink
    def test_cleanup_with_symlink_to_a_directory(self):
        # cleanup() should not follow symlinks to directories (issue #12464)
        d1 = self.do_create()
        d2 = self.do_create(recurse=0)

        # Symlink d1/foo -> d2
        os.symlink(d2.name, os.path.join(d1.name, "foo"))

        # This call to cleanup() should not follow the "foo" symlink
        d1.cleanup()

        self.assertFalse(
            os.path.exists(d1.name),
            "TemporaryDirectory %s exists after cleanup" % d1.name,
        )
        self.assertTrue(os.path.exists(d2.name), "Directory pointed to by a symlink was deleted")
        self.assertEqual(
            os.listdir(d2.name),
            ["test0.txt"],
            "Contents of the directory pointed to by a symlink " "were deleted",
        )
        d2.cleanup()

    @support.cpython_only
    def test_del_on_collection(self):
        # A TemporaryDirectory is deleted when garbage collected
        dir = tempfile.mkdtemp()
        try:
            d = self.do_create(dir=dir)
            name = d.name
            del d  # Rely on refcounting to invoke __del__
            self.assertFalse(
                os.path.exists(name),
                "TemporaryDirectory %s exists after __del__" % name,
            )
        finally:
            os.rmdir(dir)

    @support.cpython_only
    def test_del_on_collection_ignore_errors(self):
        """Test that ignoring errors works when TemporaryDirectory is gced."""
        with tempfile.TemporaryDirectory() as working_dir:
            temp_dir = self.do_create(dir=working_dir, ignore_cleanup_errors=True)
            temp_path = pathlib.Path(temp_dir.name)
            self.assertTrue(temp_path.exists(), f"TemporaryDirectory {temp_path!s} does not exist")
            with open(temp_path / "a_file.txt", "w+t") as open_file:
                open_file.write("Hello world!\n")
                del temp_dir
            self.assertEqual(
                len(list(temp_path.glob("*"))),
                int(sys.platform.startswith("win")),
                "Unexpected number of files in " f"TemporaryDirectory {temp_path!s}",
            )
            self.assertEqual(
                temp_path.exists(),
                sys.platform.startswith("win"),
                f"TemporaryDirectory {temp_path!s} existence state unexpected",
            )

    def test_del_on_shutdown(self):
        # A TemporaryDirectory may be cleaned up during shutdown
        with self.do_create() as dir:
            for mod in ("builtins", "os", "shutil", "sys", "tempfile", "warnings"):
                code = """if True:
                    import builtins
                    import os
                    import shutil
                    import sys
                    import tempfile
                    import warnings

                    tmp = tempfile.TemporaryDirectory(dir={dir!r})
                    sys.stdout.buffer.write(tmp.name.encode())

                    tmp2 = os.path.join(tmp.name, 'test_dir')
                    os.mkdir(tmp2)
                    with open(os.path.join(tmp2, "test0.txt"), "w") as f:
                        f.write("Hello world!")

                    {mod}.tmp = tmp

                    warnings.filterwarnings("always", category=ResourceWarning)
                    """.format(
                    dir=dir, mod=mod
                )
                rc, out, err = script_helper.assert_python_ok("-c", code)
                tmp_name = out.decode().strip()
                self.assertFalse(
                    os.path.exists(tmp_name),
                    "TemporaryDirectory %s exists after cleanup" % tmp_name,
                )
                err = err.decode("utf-8", "backslashreplace")
                self.assertNotIn("Exception ", err)
                self.assertIn("ResourceWarning: Implicitly cleaning up", err)

    def test_del_on_shutdown_ignore_errors(self):
        """Test ignoring errors works when a tempdir is gc'ed on shutdown."""
        with tempfile.TemporaryDirectory() as working_dir:
            code = """if True:
                import pathlib
                import sys
                import tempfile
                import warnings

                temp_dir = tempfile.TemporaryDirectory(
                    dir={working_dir!r}, ignore_cleanup_errors=True)
                sys.stdout.buffer.write(temp_dir.name.encode())

                temp_dir_2 = pathlib.Path(temp_dir.name) / "test_dir"
                temp_dir_2.mkdir()
                with open(temp_dir_2 / "test0.txt", "w") as test_file:
                    test_file.write("Hello world!")
                open_file = open(temp_dir_2 / "open_file.txt", "w")
                open_file.write("Hello world!")

                warnings.filterwarnings("always", category=ResourceWarning)
                """.format(
                working_dir=working_dir
            )
            __, out, err = script_helper.assert_python_ok("-c", code)
            temp_path = pathlib.Path(out.decode().strip())
            self.assertEqual(
                len(list(temp_path.glob("*"))),
                int(sys.platform.startswith("win")),
                "Unexpected number of files in " f"TemporaryDirectory {temp_path!s}",
            )
            self.assertEqual(
                temp_path.exists(),
                sys.platform.startswith("win"),
                f"TemporaryDirectory {temp_path!s} existence state unexpected",
            )
            err = err.decode("utf-8", "backslashreplace")
            self.assertNotIn("Exception", err)
            self.assertNotIn("Error", err)
            self.assertIn("ResourceWarning: Implicitly cleaning up", err)

    def test_exit_on_shutdown(self):
        # Issue #22427
        with self.do_create() as dir:
            code = """if True:
                import sys
                import tempfile
                import warnings

                def generator():
                    with tempfile.TemporaryDirectory(dir={dir!r}) as tmp:
                        yield tmp
                g = generator()
                sys.stdout.buffer.write(next(g).encode())

                warnings.filterwarnings("always", category=ResourceWarning)
                """.format(
                dir=dir
            )
            rc, out, err = script_helper.assert_python_ok("-c", code)
            tmp_name = out.decode().strip()
            self.assertFalse(
                os.path.exists(tmp_name),
                "TemporaryDirectory %s exists after cleanup" % tmp_name,
            )
            err = err.decode("utf-8", "backslashreplace")
            self.assertNotIn("Exception ", err)
            self.assertIn("ResourceWarning: Implicitly cleaning up", err)

    def test_warnings_on_cleanup(self):
        # ResourceWarning will be triggered by __del__
        with self.do_create() as dir:
            d = self.do_create(dir=dir, recurse=3)
            name = d.name

            # Check for the resource warning
            with warnings_helper.check_warnings(("Implicitly", ResourceWarning), quiet=False):
                warnings.filterwarnings("always", category=ResourceWarning)
                del d
                support.gc_collect()
            self.assertFalse(
                os.path.exists(name),
                "TemporaryDirectory %s exists after __del__" % name,
            )

    def test_multiple_close(self):
        # Can be cleaned-up many times without error
        d = self.do_create()
        d.cleanup()
        d.cleanup()
        d.cleanup()

    def test_context_manager(self):
        # Can be used as a context manager
        d = self.do_create()
        with d as name:
            self.assertTrue(os.path.exists(name))
            self.assertEqual(name, d.name)
        self.assertFalse(os.path.exists(name))

    def test_modes(self):
        for mode in range(8):
            mode <<= 6
            with self.subTest(mode=format(mode, "03o")):
                d = self.do_create(recurse=3, dirs=2, files=2)
                with d:
                    # Change files and directories mode recursively.
                    for root, dirs, files in os.walk(d.name, topdown=False):
                        for name in files:
                            os.chmod(os.path.join(root, name), mode)
                        os.chmod(root, mode)
                    d.cleanup()
                self.assertFalse(os.path.exists(d.name))

    @unittest.skipUnless(hasattr(os, "chflags"), "requires os.lchflags")
    def test_flags(self):
        flags = stat.UF_IMMUTABLE | stat.UF_NOUNLINK
        d = self.do_create(recurse=3, dirs=2, files=2)
        with d:
            # Change files and directories flags recursively.
            for root, dirs, files in os.walk(d.name, topdown=False):
                for name in files:
                    os.chflags(os.path.join(root, name), flags)
                os.chflags(root, flags)
            d.cleanup()
        self.assertFalse(os.path.exists(d.name))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import time
import bmy
import unittest

alt1 = "alt1"
if alt1 not in bmy.eids():
    # default using local entity
    alt1 = bmy.new(path="**/local", eid="alt1")

if not bmy.isbuilt(eid=alt1):
    bmy.build(eid=alt1)

main = "main"
if main not in bmy.eids():
    # default using local entity
    main = bmy.new(path="**/local", eid="main")

if not bmy.isbuilt(eid=main):
    bmy.build(eid=main)

bmy.select(main)

texts = ["ab", "a" * 1000]


class TestBmyExecute(unittest.TestCase):
    def test_execute_output(self):
        for text in texts:
            o, e, r = bmy.execute("echo %s" % text)
            self.assertEqual(r, 0)
            self.assertEqual(o.strip(), text)

        for text in texts:
            o, e, r = bmy.execute("echo %s >&2" % text)
            self.assertEqual(r, 0)
            self.assertEqual(e.strip(), text)

    def test_execute_exit_status(self):
        o, e, r = bmy.execute("exit 25")
        self.assertEqual(r, 25)

    def test_execute_count(self):
        count = 0
        for o, e, r in bmy.execute("echo loop", count=10):
            self.assertEqual(r, 0)
            self.assertEqual(o.strip(), "loop")
            count += 1

        self.assertEqual(count, 10)

        count = 0
        for o, e, r in bmy.execute("invalid command", count=10):
            self.assertNotEqual(r, 0)
            count += 1

        # infinite loop
        count = 0
        for o, e, r in bmy.execute("echo loop", count=-1, executein=False):
            self.assertEqual(r, 0)
            self.assertEqual(o.strip(), "loop")

            count += 1

            if count > 9:
                break

        self.assertEqual(count, 10)

    def test_execute_interval(self):
        st_time = 0
        for o, e, r in bmy.execute("echo loop", count=3, interval=5, executein=False):
            if st_time:
                end_time = time.monotonic()
                self.assertGreaterEqual(end_time - st_time, 5)
                self.assertGreaterEqual(10, end_time - st_time)
            self.assertEqual(o.strip(), "loop")
            st_time = time.monotonic()

    def test_execute_ttl(self):
        for text in texts:
            print(text)
            o, e, r = bmy.execute("echo %s" % text, ttl=10)
            self.assertEqual(r, 0)
            self.assertEqual(e.strip(), text)

    def test_execute_ttl2(self):
        exe = bmy.execute("ping 127.0.0.1", ttl=2, count=1)
        o, e, r = next(iter(exe))
        self.assertNotEqual(o.find("127.0.0.1"), -1)
